/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;

import lv.lumii.deek.style.Style;
import lv.lumii.layoutengine.*;
import lv.lumii.layoutengine.ArrangeData.ArrangeStyle;
import lv.lumii.layoutengine.Box;
import lv.lumii.layoutengine.Container;
import lv.lumii.layoutengine.LayoutConstraints.ConstraintType;
import lv.lumii.layoutengine.LayoutConstraints.GridLayoutConstraints;
import lv.lumii.deek.util.Globals;
import org.w3c.dom.*;
import org.w3c.dom.Element;

import javax.swing.*;

/**
 * Abstract class describing two-dimensional diagram shapes.
 * <p/>
 * Containers can have an arrange style associated with them. Arrange style determines how the diagram elements are layouted.
 * <p/>
 * Containers can have a grid for hosting its children.
 */
public abstract class DContainer extends DElement {

DContainer(lv.lumii.layoutengine.Container container) {
    super(container);
}

DContainer(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    super(e, objMap);
}

@Override
public lv.lumii.layoutengine.Container getLayoutElement() {
    return (lv.lumii.layoutengine.Container) layoutElement;
}

@Override
public DContainer getOwner() {
    // TODO better
    return (DContainer) getThis(layoutElement.getOwner());
}

@Override
protected void setupElementStyle(Style style) {

}

@Override
protected void drawSelf(Graphics2D g) {
}

@Override
protected void drawDecorations(Graphics2D g) {
    drawGrid(g);
}

private void drawGrid(Graphics2D g) {

    lv.lumii.layoutengine.Container lel = getLayoutElement();

    if (lel.getLayoutConstraints().getType() == ConstraintType.GRID) {

        Rectangle2D.Double r;
        if (this instanceof DBox) {
            r = getBounds();
        } else { // group or diagram
//            r = lel.findDescendantEnclosingRectangle();
            Path2D.Double shape = lel.getShape();
            r = (Rectangle2D.Double) (shape.getBounds2D());
        }

        if (r == null)
            return;

        GridLayoutConstraints glc = (GridLayoutConstraints) lel.getLayoutConstraints();

        g.setColor(Globals.GRID_BORDER_COLOR);

        if (glc.getColumnCount() > 1) {
            for (int i = 2; i <= glc.getColumnCount(); i++) {
                g.draw(new Line2D.Double(glc.getColumnLeft(i), r.y, glc.getColumnLeft(i), r.y + r.height));
            }
        }

        if (glc.getRowCount() > 1) {
            for (int i = 2; i <= glc.getRowCount(); i++) {
                g.draw(new Line2D.Double(r.x, glc.getRowTop(i), r.x + r.width, glc.getRowTop(i)));
            }

        }
    }

}

@Override
protected void drawChildren(Graphics2D g) {

    lv.lumii.layoutengine.Container lel = getLayoutElement();

    Rectangle clipRect = g.getClipBounds(); // TODO

    if (clipRect != null) {

        for (DBox e : boxes()) {
            if (e.getBounds().intersects(clipRect))
                e.draw(g);
        }

        for (DGroup e : groups()) {
//                if (e.getBounds().intersects(clipRect))
            e.draw(g);
        }

        for (DInsideLabel e : insideLabels()) {
            if (e.getBounds().intersects(clipRect))
                e.draw(g);
        }

        for (DLine e : lines()) {
//                if (e.getBounds().intersects(clipRect))
            e.draw(g);
        }

    } else {

        for (DBox e : boxes()) {
            e.draw(g);
        }

        for (DGroup e : groups()) {
            e.draw(g);
        }

        for (DInsideLabel e : insideLabels()) {
            e.draw(g);
        }

        for (DLine e : lines()) {
            e.draw(g);
        }

    }


}

@Override
public void removeChildren() {
    for (DLine l : lines())
        l.remove();

    for (DBox b : boxes())
        b.remove();

    for (DGroup g : groups())
        g.remove();

    for (DInsideLabel e : insideLabels())
        e.remove();
}

protected Color getArrangeStyleColor() {
    ArrangeStyle as = getLayoutElement().getArrangeStyle();

    switch (as) {
        case FLOW:
            return Globals.AS_FLOW_FILL_COLOR;

        case SPRING_EMBEDDED:
            return Globals.AS_SPRING_FILL_COLOR;

        case INHERITED:
            return Globals.AS_INHERITED_FILL_COLOR;

        case UNIVERSAL:
            return Globals.AS_UNIVERSAL_FILL_COLOR;

        case NONE:
        default:
            return Globals.AS_NONE_FILL_COLOR;
    }
}

@Override
protected String getTestInfoString() {

    lv.lumii.layoutengine.Container lc = getLayoutElement();

    String thisHostInfo = "";
    String thisGuestInfo = "";

    GridLayoutConstraints glc;

    lv.lumii.layoutengine.Container parent = lc.getOwner();

    // grid host
    if (lc.getLayoutConstraints().getType() == ConstraintType.GRID) {
        glc = (GridLayoutConstraints) lc.getLayoutConstraints();
        thisHostInfo = String.format("[%d,%d]",
                glc.getRowCount(),
                glc.getColumnCount());

    }

    // grid guest
    if (parent.getLayoutConstraints().getType() == ConstraintType.GRID) {
        glc = (GridLayoutConstraints) parent.getLayoutConstraints();
        thisGuestInfo = String.format("{%s,%s}",
                glc.getRow(lc),
                glc.getColumn(lc));
    }

    String containerInfo = String.format("%s %s %s %s",
            lc.getArrangeStyle(),
            lc.getLayoutConstraints().getType(),
            thisHostInfo,
            thisGuestInfo);

    return super.getTestInfoString() + containerInfo;
}

@Override
public boolean isSelfHit(Point2D.Double p) {
    return false;
}

@Override
public boolean isChildHit(Point2D.Double p) {

    for (DLine e : lines()) {
        if (e.isHit(p))
            return true;
    }

    for (DGroup e : groups()) {
        if (e.isHit(p))
            return true;
    }

    for (DBox e : boxesReverse()) {
        if (e.isHit(p))
            return true;
    }

    for (DInsideLabel e : insideLabels()) {
        if (e.isHit(p))
            return true;
    }

    return false;
}

@Override
public DElement getChildHit(Point2D.Double p) {
    for (DInsideLabel e : insideLabels()) {
        DElement h = e.getHit(p);
        if (h != null)
            return h;
    }

    for (DLine e : lines()) {
        DElement h = e.getHit(p);
        if (h != null)
            return h;
    }

    for (DGroup e : groups()) {
        DElement h = e.getHit(p);
        if (h != null)
            return h;
    }

    for (DBox e : boxesReverse()) {
        DElement h = e.getHit(p);
        if (h != null)
            return h;
    }

    return null;
}

public DBox getHitBox(Point2D.Double p) {

    for (DBox b : boxes()) {
        DBox b2 = b.getHitBox(p);
        if (b2 != null)
            return b2;
    }

    for (DGroup g : groups()) {
        DBox b2 = g.getHitBox(p);
        if (b2 != null)
            return b2;
    }

    return null;
}

public Iterable<DBox> boxes() {
    return new Iterable<DBox>() {
        @Override
        public Iterator<DBox> iterator() {
            return new Iterator<DBox>() {
                private ArrayList<Box> _items = getLayoutElement().getBoxes();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DBox next() {
                    return getThisBox(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DBox> boxesReverse() {
    return new Iterable<DBox>() {
        @Override
        public Iterator<DBox> iterator() {
            return new Iterator<DBox>() {
                private ArrayList<Box> _items = getLayoutElement().getBoxes();
                private int pos = _items.size();

                @Override
                public boolean hasNext() {
                    return pos > 0;
                }

                @Override
                public DBox next() {
                    return getThisBox(_items.get(--pos));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DGroup> groups() {
    return new Iterable<DGroup>() {
        @Override
        public Iterator<DGroup> iterator() {
            return new Iterator<DGroup>() {
                private ArrayList<Container> _items = getLayoutElement().getPureContainers();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DGroup next() {
                    return DGroup.getThisGroup(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DContainer> containers() {
    return new Iterable<DContainer>() {
        @Override
        public Iterator<DContainer> iterator() {
            return new Iterator<DContainer>() {
                private ArrayList<Container> _items = getLayoutElement().getContainers();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DContainer next() {
                    return DContainer.getThisContainer(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DContainer> allContainers() {
    return new Iterable<DContainer>() {
        @Override
        public Iterator<DContainer> iterator() {
            return new Iterator<DContainer>() {
                private ArrayList<Container> _items = getLayoutElement().getDescendantContainers();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DContainer next() {
                    return DContainer.getThisContainer(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DLine> lines() {
    return new Iterable<DLine>() {
        @Override
        public Iterator<DLine> iterator() {
            return new Iterator<DLine>() {
                private ArrayList<Line> _items = getLayoutElement().getLines();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DLine next() {
                    return DLine.getThisLine(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DInsideLabel> insideLabels() {
    return new Iterable<DInsideLabel>() {
        @Override
        public Iterator<DInsideLabel> iterator() {
            return new Iterator<DInsideLabel>() {
                private ArrayList<InsideLabel> _items = getLayoutElement().getInsideLabels();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DInsideLabel next() {
                    return DElement.getThisInsideLabel(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DBox> allBoxes() {
    return new Iterable<DBox>() {
        @Override
        public Iterator<DBox> iterator() {
            return new Iterator<DBox>() {
                private ArrayList<Box> _items = getLayoutElement().getDescendantBoxes();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DBox next() {
                    return getThisBox(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DLine> allLines() {
    return new Iterable<DLine>() {
        @Override
        public Iterator<DLine> iterator() {
            return new Iterator<DLine>() {
                private ArrayList<Line> _layoutLines = getLayoutElement().getDescendantLines();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _layoutLines.size();
                }

                @Override
                public DLine next() {
                    return DElement.getThisLine(_layoutLines.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public Iterable<DLine> allLinesReverse() {
    return new Iterable<DLine>() {
        @Override
        public Iterator<DLine> iterator() {
            return new Iterator<DLine>() {
                private ArrayList<Line> _layoutLines = getLayoutElement().getDescendantLines();
                private int pos = _layoutLines.size();

                @Override
                public boolean hasNext() {
                    return pos > 0;
                }

                @Override
                public DLine next() {
                    return DLine.getThisLine(_layoutLines.get(--pos));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

public ArrayList<DBox> getAllBoxes() {
    ArrayList<DBox> all = new ArrayList<>();
    for (DBox e : allBoxes()) {
        all.add(e);
    }
    return all;
}

public ArrayList<DLine> getAllLines() {
    ArrayList<DLine> all = new ArrayList<>();
    for (DLine e : allLines()) {
        all.add(e);
    }
    return all;
}

public ArrangeStyle getArrangeStyle() {
    return getLayoutElement().getArrangeStyle();
}

public void setArrangeStyle(ArrangeStyle as) {
    // TODO - call this for arrange style changes

    getLayoutElement().setArrangeStyle(as);
    style.setFillColor(getArrangeStyleColor());
    updateAll();
}

public ArrangeData getArrangeData() {
    return getLayoutElement().getArrangeData();
}

public LayoutConstraints getLayoutConstraints() {
    return getLayoutElement().getLayoutConstraints();
}

public void setConstraintType(ConstraintType constraintType) {
    getLayoutElement().setConstraintType(constraintType);
}

/**
 * Convert this container to a box; all children will be preserved.
 *
 * @return this container ar a box
 */
public abstract DBox convertToBox();

/**
 * Convert this container to a group; lines to/from this box will be dropped, all other children will be preserved.
 *
 * @return this container as a group
 */
public abstract DGroup convertToGroup();

/**
 * Convert this container to a group; lines to/from this box will be dropped, all other children will be preserved.
 *
 * @param moveVector
 * @param owner
 * @return
 */
abstract DGroup convertToGroup(Point2D.Double moveVector, DContainer owner);

/**
 * Convert this container to a group; lines to/from this box will be dropped, all other children will be preserved.
 *
 * @param moveVector
 * @param owner
 * @param gridRow
 * @param gridColumn
 * @return
 */
abstract DGroup convertToGroup(Point2D.Double moveVector, DContainer owner, Integer gridRow, Integer gridColumn);

/**
 * Move this container by a vector / to a new owner
 *
 * @param moveVector the move vector
 * @param owner      the new owner
 */
public void move(Point2D.Double moveVector, DContainer owner) {
    getLayoutElement().move(moveVector, owner.getLayoutElement());
}

/**
 * Move this container by a vector / to a new owner
 *
 * @param moveVector the move vector
 * @param owner      the new owner
 * @param gridRow    grid row for this label (box parent should have a grid)
 * @param gridColumn grid column for this label (box parent should have a grid)
 */
public void move(Point2D.Double moveVector, DContainer owner, Integer gridRow, Integer gridColumn) {
    getLayoutElement().move(moveVector, owner.getLayoutElement(), gridRow, gridColumn);
}

/**
 * @param newCenter  the desired center for the new box
 * @param owner      the new owner
 * @param growPoint  the grow point
 * @param gridRow    grid row for this label (box parent should have a grid)
 * @param gridColumn grid column for this label (box parent should have a grid)
 * @return converted box
 */
public DBox convertToBox(Point2D.Double newCenter, DContainer owner, Point2D.Double growPoint, Integer gridRow, Integer gridColumn) {
    Box c = getLayoutElement().convertToBox(newCenter, owner.getLayoutElement(), growPoint, gridRow, gridColumn);
    return new DBox(c);
}

/**
 * @param newCenter the desired center for the new box
 * @param owner     the new owner
 * @param growPoint the grow point
 * @return converted box
 */
public DBox convertToBox(Point2D.Double newCenter, DContainer owner, Point2D.Double growPoint) {
    Box c = getLayoutElement().convertToBox(newCenter, owner.getLayoutElement(), growPoint);
    return new DBox(c);
}

/**
 * Create new inside label
 *
 * @param newRect           rectangle providing the label shape
 * @param constraintType    constraint type for this label
 * @param currentBoxSpacing spacing for this label
 * @param gridRow           grid row for this label (label's parent should have a grid)
 * @param gridColumn        grid column for this label (label's parent should have a grid)
 * @return the new label
 */
public DInsideLabel createInsideLabel(Rectangle2D.Double newRect, ConstraintType constraintType, double currentBoxSpacing, Integer gridRow, Integer gridColumn) {
    return new DInsideLabel(getLayoutElement().createInsideLabel(newRect, constraintType, currentBoxSpacing, gridRow, gridColumn));
}

/**
 * Create new inside label
 *
 * @param newRect           rectangle providing the label shape
 * @param constraintType    constraint type for this label
 * @param currentBoxSpacing spacing for this label
 * @return the new label
 */
public DInsideLabel createInsideLabel(Rectangle2D.Double newRect, ConstraintType constraintType, double currentBoxSpacing) {
    return new DInsideLabel(getLayoutElement().createInsideLabel(newRect, constraintType, currentBoxSpacing));
}

/**
 * Create new inside label
 *
 * @param newRect        rectangle providing the label shape
 * @param constraintType constraint type for this label
 * @return the new label
 */
public DInsideLabel createInsideLabel(Rectangle2D.Double newRect, ConstraintType constraintType) {
    return new DInsideLabel(getLayoutElement().createInsideLabel(newRect, constraintType, Globals.getCurrentBoxSpacing()));
}

/**
 * Create new inside label
 *
 * @param newRect rectangle providing the label shape
 * @return the new label
 */
public DInsideLabel createInsideLabel(Rectangle2D.Double newRect) {
    return new DInsideLabel(getLayoutElement().createInsideLabel(newRect, ConstraintType.NONE, Globals.getCurrentBoxSpacing()));
}

/**
 * Create a new group
 *
 * @return the group created
 */
public DGroup createGroup() {
    return new DGroup(getLayoutElement().createContainer());
}

/**
 * Create a new group with the provided arrange style and constraint type.
 *
 * @param arrangeStyle   arrange style for the group
 * @param constraintType constraint type for the group
 * @return the group created
 */
public DGroup createGroup(ArrangeStyle arrangeStyle, ConstraintType constraintType) {
    return new DGroup(getLayoutElement().createContainer(arrangeStyle, constraintType));
}

/**
 * Create a new box without specific position and size.
 *
 * @return the box created
 */
public DBox createBox() {
    return createBox(new Rectangle2D.Double());
}

/**
 * Create a new box with the given geometry described by a rectangle.
 *
 * @param newRect rectangle describing the box geometry
 * @return the box created
 */
public DBox createBox(Rectangle2D.Double newRect) {
    return createBox(newRect, ArrangeStyle.INHERITED, ConstraintType.NONE, Globals.getCurrentBoxSpacing());
}

/**
 * @param newRect           rectangle describing the box geometry
 * @param arrangeStyle      arrange style for the new box
 * @param constraintType    constraint type for the new box
 * @param currentBoxSpacing spacing for the new box
 * @return the box created
 */
public DBox createBox(Rectangle2D.Double newRect, ArrangeStyle arrangeStyle, ConstraintType constraintType, double currentBoxSpacing) {
    return new DBox(getLayoutElement().createBox(newRect, arrangeStyle, constraintType, currentBoxSpacing));
}

/**
 * @param newRect           rectangle describing the box geometry
 * @param arrangeStyle      arrange style for the new box
 * @param constraintType    constraint type for the new box
 * @param currentBoxSpacing spacing for the new box
 * @param gridRow           grid row for the box created (box parent should have a grid)
 * @param gridColumn        grid column for the box created (box parent should have a grid)
 * @return the box created
 */
public DBox createBox(Rectangle2D.Double newRect, ArrangeStyle arrangeStyle, ConstraintType constraintType, double currentBoxSpacing, Integer gridRow, Integer gridColumn) {
    return new DBox(getLayoutElement().createBox(newRect, arrangeStyle, constraintType, currentBoxSpacing, gridRow, gridColumn));
}

/**
 * Returns the context menu for this container.
 *
 * @return the context menu
 */
@Override
public JPopupMenu getElementMenu() {
    JPopupMenu menu = super.getElementMenu();

    addHostMenuItems(menu);

    if (getOwner() != null) {
        addGuestMenuItems(menu);
    }

    //...

    return menu;
}

private void addGuestMenuItems(JPopupMenu menu) {

    DContainer host = getOwner();
    if (host.getLayoutConstraints().getType() == ConstraintType.GRID) {

        menu.addSeparator();

        GridLayoutConstraints glc = (GridLayoutConstraints) host.getLayoutConstraints();

        int rowNumber;
        if (glc.getRow(getLayoutElement()) != null) {
            rowNumber = glc.getRow(getLayoutElement());
        } else {
            rowNumber = Commands.GRID_ROWCOL_NONE_INTVALUE;
        }
        menu.add(getNumberSubmenu(Commands.GRID_SET_ROW_MENU, Commands.GRID_SET_ROW_PREFIX,
                glc.getRowCount(),
                rowNumber,
                true,
                new ActionListener() {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        String actionCommand = e.getActionCommand();
                        if (actionCommand.equals(Commands.GRID_ROWCOL_NONE_STRINGVALUE))
                            setGridRow(null);
                        else
                            setGridRow(Integer.parseInt(actionCommand));
                        updateAll();
                    }
                }));

        int colNumber;
        if (glc.getColumn(getLayoutElement()) != null) {
            colNumber = glc.getColumn(getLayoutElement());
        } else {
            colNumber = Commands.GRID_ROWCOL_NONE_INTVALUE;
        }
        menu.add(getNumberSubmenu(Commands.GRID_SET_COL_MENU, Commands.GRID_SET_COL_PREFIX,
                glc.getColumnCount(),
                colNumber,
                true,
                new ActionListener() {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        String actionCommand = e.getActionCommand();
                        if (actionCommand.equals(Commands.GRID_ROWCOL_NONE_STRINGVALUE))
                            setGridColumn(null);
                        else
                            setGridColumn(Integer.parseInt(actionCommand));
                        updateAll();
                    }
                }));

    }
}

protected void addHostMenuItems(JPopupMenu menu) {

    menu.addSeparator();

    menu.add(getArrangeSubmenu(getArrangeStyle()));

    if (getArrangeStyle() == ArrangeStyle.FLOW) {
        ArrangeData.FlowData flowData = (ArrangeData.FlowData) (getArrangeData());
        menu.add(getFlowDirectionSubmenu(flowData.getFlowDirection()));
    }

    menu.addSeparator();

    menu.add(getConstraintTypeSubmenu(getLayoutConstraints().getType()));

    if (getLayoutConstraints().getType() == ConstraintType.GRID) {

        final GridLayoutConstraints glc = (GridLayoutConstraints) getLayoutConstraints();

        menu.add(getNumberSubmenu(Commands.GRID_SET_ROWCOUNT_MENU, Commands.GRID_SET_ROWCOUNT_PREFIX,
                Commands.GRID_MAXROWS, glc.getRowCount(), false, new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                String actionCommand = e.getActionCommand();
                int newCount = Integer.parseInt(actionCommand);
                if (glc.getRowCount() < newCount)
                    glc.addRows(newCount - glc.getRowCount());
            }
        }));

        menu.add(getNumberSubmenu(Commands.GRID_SET_COLCOUNT_MENU, Commands.GRID_SET_COLCOUNT_PREFIX,
                Commands.GRID_MAXCOLS, glc.getColumnCount(), false, new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                String actionCommand = e.getActionCommand();
                int newCount = Integer.parseInt(actionCommand);
                if (glc.getColumnCount() < newCount)
                    glc.addColumns(newCount - glc.getColumnCount());
            }
        }));

    }
}

protected JMenu getArrangeSubmenu(ArrangeStyle currentStyle) {
    JMenu arrangeSubmenu = new JMenu(Commands.ARRANGE_STYLE_MENU);
    JMenuItem item;

    for (final ArrangeStyle as : ArrangeStyle.values()) {
        item = new JCheckBoxMenuItem(as.toString());
        if (currentStyle == as)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setArrangeStyle(as);
                getDiagram().arrange();
            }
        });
        item.setActionCommand(Commands.ARRANGE_STYLE_PREFIX + ':' + as.toString());
        arrangeSubmenu.add(item);
    }
    if (currentStyle == ArrangeStyle.FLOW) {

    }
    return arrangeSubmenu;
}

private JMenu getFlowDirectionSubmenu(ArrangeData.FlowData.Direction currentDirection) {

    JMenuItem item;
    JMenu directionSubmenu = new JMenu(Commands.ARRANGE_STYLE_FLOW_DIRECTION_MENU);

    for (final ArrangeData.FlowData.Direction fd : ArrangeData.FlowData.Direction.values()) {
        item = new JCheckBoxMenuItem(fd.toString());
        if (currentDirection == fd)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                // ignore all but flow
                if (getArrangeStyle() == ArrangeStyle.FLOW) {
                    ArrangeData.FlowData flowData = (ArrangeData.FlowData) getArrangeData();
                    if (flowData.getFlowDirection() != fd) {
                        flowData.setFlowDirection(fd);

                        updateAll();
                    }
                }
            }
        });
        item.setActionCommand(Commands.ARRANGE_STYLE_FLOW_DIRECTION_PREFIX + ':' + fd.toString());
        directionSubmenu.add(item);
    }

    return directionSubmenu;
}

private JMenu getConstraintTypeSubmenu(ConstraintType currentType) {

    JMenuItem item;
    JMenu constraintSubmenu = new JMenu(Commands.CONSTRAINT_TYPE_MENU);

    for (final ConstraintType ct : ConstraintType.values()) {
        item = new JCheckBoxMenuItem(ct.toString());
        if (currentType == ct)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                ConstraintType ct0 = getLayoutElement().getLayoutConstraints().getType();
                if (ct != ct0) {
                    setConstraintType(ct);
                    if (ct == ConstraintType.GRID) {
                        GridLayoutConstraints glc = (GridLayoutConstraints) getLayoutConstraints();
                        glc.addColumns(2);
                        glc.addRows(1);
                    }

                    updateAll();
                }

            }
        });
        item.setActionCommand(Commands.CONSTRAINT_TYPE_PREFIX + ':' + ct.toString());
        constraintSubmenu.add(item);
    }
    return constraintSubmenu;
}

private JMenu getNumberSubmenu(String menuName, String menuPrefix, int maxNumber, int currentNumber, boolean allowNull, ActionListener listener) {

    JMenuItem item;
    JMenu numberSubmenu = new JMenu(menuName);

    for (int i = 1; i <= maxNumber; i++) {
        item = new JCheckBoxMenuItem(Integer.toString(i));
        if (i == currentNumber)
            item.setSelected(true);
        item.addActionListener(listener);
        item.setActionCommand(Integer.toString(i));
        numberSubmenu.add(item);
    }

    if (allowNull) {
        item = new JCheckBoxMenuItem(Commands.GRID_ROWCOL_NONE);
        if (Commands.GRID_ROWCOL_NONE_INTVALUE == currentNumber)
            item.setSelected(true);
        item.addActionListener(listener);
        item.setActionCommand(Commands.GRID_ROWCOL_NONE_STRINGVALUE);
        numberSubmenu.add(item);
    }

    return numberSubmenu;

}

Point2D.Double getCenter() {
    return new Point2D.Double(getBounds().getCenterX(), getBounds().getCenterY());
}

@Override
protected void saveChildrenToXML(Element e, Document doc, LinkedHashMap<Object, Integer> idMap) {
    super.saveChildrenToXML(e, doc, idMap);

    Element eGroups = doc.createElement(XMLHelper.GROUPS_ELEMENT);
    for (DGroup group : groups()) {
        eGroups.appendChild(group.saveToXML(doc, idMap));
    }
    e.appendChild(eGroups);

    Element eBoxes = doc.createElement(XMLHelper.BOXES_ELEMENT);
    for (DBox box : boxes()) {
        eBoxes.appendChild(box.saveToXML(doc, idMap));
    }
    e.appendChild(eBoxes);

    Element eLines = doc.createElement(XMLHelper.LINES_ELEMENT);
    for (DLine line : lines()) {
        eLines.appendChild(line.saveToXML(doc, idMap));
    }
    e.appendChild(eLines);

    Element eInsideLabels = doc.createElement(XMLHelper.INSIDE_LABELS_ELEMENT);
    for (DInsideLabel lbl : insideLabels()) {
        eInsideLabels.appendChild(lbl.saveToXML(doc, idMap));
    }
    e.appendChild(eInsideLabels);

}

private static class Commands {
    // do not use ':' in commands and prefixes, as it is used as separator here

    static final String ARRANGE_STYLE_MENU = "Arrange Style";
    static final String ARRANGE_STYLE_PREFIX = "arrangestyle";
    static final String ARRANGE_STYLE_FLOW_DIRECTION_MENU = "Flow Direction";
    static final String ARRANGE_STYLE_FLOW_DIRECTION_PREFIX = "flowdirection";
    static final String CONSTRAINT_TYPE_MENU = "Constraint Type";
    static final String CONSTRAINT_TYPE_PREFIX = "constrainttype";

    //		static final String GRID_SET_DIMENSIONS = "Set grid dimensions...";
//		static final String GRID_SET_DIMENSIONS_COMMAND = "constraints-grid-setdimensions";
    static final String GRID_SET_ROWCOUNT_MENU = "Row count";
    static final String GRID_SET_ROWCOUNT_PREFIX = "grid-rowcount";
    static final String GRID_SET_COLCOUNT_MENU = "Column count";
    static final String GRID_SET_COLCOUNT_PREFIX = "grid-colcount";
    static final int GRID_MAXROWS = 5;
    static final int GRID_MAXCOLS = 7;
    static final String GRID_SET_ROW_MENU = "Set row";
    static final String GRID_SET_ROW_PREFIX = "grid-row";
    static final String GRID_SET_COL_MENU = "Set column";
    static final String GRID_SET_COL_PREFIX = "grid-col";
    static final String GRID_ROWCOL_NONE = "none";
    static final String GRID_ROWCOL_NONE_STRINGVALUE = "none";
    static final int GRID_ROWCOL_NONE_INTVALUE = -1;
    static final String GRIDIFY_CHILDREN = "Gridify children";
    static final String GRIDIFY_CHILDREN_COMMAND = "constraints-grid-gridifychildren";
    static final String LABELIFY_CHILDREN = "Child boxes to labels";
    static final String LABELIFY_CHILDREN_COMMAND = "test-labelifychildren";
    static final String CONVERT_TO_BOX = "Convert to box";
    static final String CONVERT_TO_BOX_COMMAND = "group-converttobox";
    static final String CONVERT_TO_GROUP = "Convert to group";
    static final String CONVERT_TO_GROUP_COMMAND = "group-converttogroup";
    static final String ARRANGE_DIAGRAM = "Arrange Diagram";
    static final String ARRANGE_DIAGRAM_COMMAND = "arrange-_drawing";
    static final String MANUAL_MODE = "Manual mode";
    static final String MANUAL_MODE_COMMAND = "toggle-manual";
    static final String AUTO_ARRANGE = "Auto-Arrange";
    static final String AUTO_ARRANGE_COMMAND = "autoarrange";
    static final String SIMPLIFY_ALL_LINES = "Simplify all lines";
    static final String SIMPLIFY_ALL_LINES_COMMAND = "lines-simplify";
    static final String AUTO_SIMPLIFY = "Auto-Simplify";
    static final String AUTO_SIMPLIFY_COMMAND = "autosimplify";
    static final String RETRACE_ALL_LINES = "Retrace all lines";
    static final String RETRACE_ALL_LINES_COMMAND = "lines-retrace";
    static final String LOAD_DIAGRAM_BACKGROUND = "Load Background ...";
    static final String LOAD_DIAGRAM_BACKGROUND_COMMAND = "background-load";
    static final String TOGGLE_DIAGRAM_BACKGROUND_VISIBLE = "Background visible";
    static final String TOGGLE_DIAGRAM_BACKGROUND_VISIBLE_COMMAND = "background-toggle";
    static final String RECOGNIZE_DIAGRAM_BACKGROUND = "Recognize Background";
    static final String RECOGNIZE_DIAGRAM_BACKGROUND_COMMAND = "background-recognize";
    static final String RECOGNIZE_CURRENT_DIAGRAM = "Recognize Current";
    static final String RECOGNIZE_CURRENT_DIAGRAM_COMMAND = "background-setcurrentandrecognize";
    static final String DIAGRAM_LINE_TYPE_MENU = "Diagram line type";
    static final String DIAGRAM_LINE_TYPE_PREFIX = "_drawing-setlinetype";
    static final String BOX_INCOMING_LINE_SIDES_MENU = "Incoming line sides";
    static final String BOX_INCOMING_LINE_SIDES_PREFIX = "boxlines-setincomingsides";
    static final String BOX_OUTGOING_LINE_SIDES_MENU = "Outgoing line sides";
    static final String BOX_OUTGOING_LINE_SIDES_PREFIX = "boxlines-setoutgoingsides";
    static final String BOX_RESET_LINE_SIDES = "Reset sides";
    static final String BOX_RESET_LINE_SIDES_COMMAND = "RESET";
}
}
