/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;

import lv.lumii.deek.style.Style;
import lv.lumii.deek.ui.DDrawing;
import lv.lumii.layoutengine.LayoutConstraints.ConstraintType;
import lv.lumii.layoutengine.LayoutConstraints.GridLayoutConstraints;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.swing.*;

/**
 * The main container for all elements in a diagram.
 * <p/>
 * Diagram is similar to a group, but additionally has some specific, global operations, which affect all elements in the diagram..
 */
public class DDiagram extends DContainer {

private DDrawing _drawing;
private float _shadowOpacity = 0.3f;
private double _shadowOffsetX = 3;
private double _shadowOffsetY = -3;
private Color _shadowColor = new Color(0x000000);


public DDiagram(lv.lumii.layoutengine.Diagram layout) {
    super(layout);
    initDiagram();
}

public DDiagram(lv.lumii.layoutengine.Diagram layout, DDrawing drawing) {
    this(layout);
    this._drawing = drawing;
}

public DDiagram(org.w3c.dom.Element diagramElement, LinkedHashMap<Integer, Object> objectMap,
                lv.lumii.layoutengine.Diagram newLayout, DDrawing drawing) {
    this(diagramElement, objectMap, newLayout);
    this._drawing = drawing;
}

public DDiagram(org.w3c.dom.Element diagramElement, LinkedHashMap<Integer, Object> objectMap,
                lv.lumii.layoutengine.Diagram newLayout) {

    super(diagramElement, objectMap);

}

@Override
public lv.lumii.layoutengine.Diagram getLayoutElement() {
    return (lv.lumii.layoutengine.Diagram) layoutElement;
}

@Override
protected void setupElementStyle(Style style) {
    style.setFillColor(new Color(255, 255, 208));
}

private void initDiagram() {

}

@Override
public DDrawing getDrawing() {
    return _drawing;
}

void setDrawing(DDrawing d) {
    _drawing = d;
}

@Override
public Rectangle2D.Double getBounds() {
    return getLayoutElement().getBounds();
}

public boolean isManualMode() {
    if (getLayoutElement().getState() == lv.lumii.layoutengine.Diagram.State.MANUAL)
        return true;

    return false;
}

public void setManualMode(boolean manual) {

    if (manual && isManualMode()) return;
    if (!manual && !isManualMode()) return;

    if (manual) {
        getLayoutElement().startManual();
    } else {
        getLayoutElement().endManual();
    }

}

@Override
public DBox convertToBox() {
    throw new UnsupportedOperationException("Diagram cannot be converted to a box");
}

@Override
public DGroup convertToGroup() {
    throw new UnsupportedOperationException("Diagram cannot be converted to a group");
}

@Override
DGroup convertToGroup(Point2D.Double moveVector, DContainer owner) {
    return null;
}

@Override
DGroup convertToGroup(Point2D.Double moveVector, DContainer owner, Integer gridRow, Integer gridColumn) {
    return null;
}

@Override
protected void drawSelf(Graphics2D g) {

    Rectangle2D.Double bounds = getLayoutElement().getBounds();
    if (bounds != null) {
        g.setColor(style.getFillColor());
        Rectangle2D erase = (g.getClipBounds() != null)
                ? bounds.createIntersection(g.getClipBounds())
                : bounds;
        g.fill(erase);
    }

}

@Override
public boolean isHit(Point2D.Double p) {
    return false;
}

@Override
public void recalculate() {

}

protected void drawGrid(Graphics2D g) {
    if (getLayoutElement().getLayoutConstraints().getType() == ConstraintType.GRID) {
        GridLayoutConstraints glc = (GridLayoutConstraints) getLayoutElement().getLayoutConstraints();

        Rectangle2D.Double cellRect;

        Paint evenCellPaint = getArrangeStyleColor();
        Paint oddCellPaint = getArrangeStyleColor().brighter();

        for (int iRow = 1; iRow <= glc.getRowCount(); iRow++) {
            for (int iColumn = 1; iColumn <= glc.getColumnCount(); iColumn++) {

                cellRect = new Rectangle2D.Double(
                        glc.getColumnLeft(iColumn),
                        glc.getRowTop(iRow),
                        glc.getColumnWidth(iColumn),
                        glc.getRowHeight(iRow)
                );

                g.setPaint((iColumn + iRow) % 2 == 0 ? evenCellPaint : oddCellPaint);
                g.fill(cellRect);

            }
        }

    }

}

boolean usesShadow() {
    return true;
}

void useShadow(boolean useIt) {
}

double getShadowOpacityy() {
    return _shadowOpacity;
}

double getShadowOffsetX() {
    return _shadowOffsetX;
}

double getShadowOffsetY() {
    return _shadowOffsetY;
}

Paint getShadowPaint() {
    return Color.GRAY;
}

Color getShadowColor() {
    return Color.GRAY;
}

public void arrange() {
    getLayoutElement().arrange();
}

@Override
public void startTransaction() {
    getLayoutElement().startTransaction();
}

@Override
public void endTransaction() {
    getLayoutElement().endTransaction();
}

@Override
public JPopupMenu getElementMenu() {

    JPopupMenu menu = new JPopupMenu();

    menu.add(menuCommand(Commands.ARRANGE_DIAGRAM, Commands.ARRANGE_DIAGRAM_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            getDrawing().arrangeDiagram();
            getDrawing().zoomToContents();
        }
    }));

    menu.add(menuCommand(Commands.AUTO_ARRANGE, Commands.AUTO_ARRANGE_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            getDrawing().setAutoArrange(!getDrawing().hasAutoArrange());
            updateAll();
        }
    },
            getDrawing().hasAutoArrange()
    ));

    menu.add(menuCommand(Commands.MANUAL_MODE, Commands.MANUAL_MODE_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            if (getDrawing().isManualMode()) {
                getDrawing().startTransaction("exit manual mode");
                getDrawing().setManualMode(false);
                getDrawing().endTransaction();
            } else {
                getDrawing().setManualMode(true);

            }
        }
    },
            getDrawing().isManualMode()
    ));

    addHostMenuItems(menu);

    menu.addSeparator();

    menu.add(getLineTypeSubmenu(Commands.DIAGRAM_LINE_TYPE_MENU, Commands.DIAGRAM_LINE_TYPE_PREFIX, getDrawing().getCurrentLineType(), new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            String actionCommand = e.getActionCommand();
            getDrawing().setCurrentLineType(DLine.DLineType.valueOf(actionCommand));
            updateAll();
        }
    }));

    menu.add(menuCommand(Commands.SIMPLIFY_ALL_LINES, Commands.SIMPLIFY_ALL_LINES_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            getDrawing().cleanupAllOrthoLines();
            updateAll();
        }
    }));

    menu.add(menuCommand(Commands.AUTO_SIMPLIFY, Commands.AUTO_SIMPLIFY_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            getDrawing().setLineAutoCleanup(!getDrawing().hasLineAutoCleanup());
        }
    },
            getDrawing().hasLineAutoCleanup()
    ));

    menu.add(menuCommand(Commands.RETRACE_ALL_LINES, Commands.RETRACE_ALL_LINES_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            getDrawing().retraceAllLines();
            updateAll();
        }
    }));


    //...

    return menu;
}

public org.w3c.dom.Element saveToXML(org.w3c.dom.Document doc) {

    LinkedHashMap<Object, Integer> idMap = new LinkedHashMap<>();
    org.w3c.dom.Element layoutDOM = getLayoutElement().saveToXML(doc, idMap);

    org.w3c.dom.Element e = saveToXML(doc, idMap);

    e.appendChild(layoutDOM);

    return e;
}

@Override
void saveAttributesToXML(Element e) {
    super.saveAttributesToXML(e);
}

private static class Commands {
    // do not use ':' in commands and prefixes, as it is used as separator here

    static final String ARRANGE_DIAGRAM = "Arrange Diagram";
    static final String ARRANGE_DIAGRAM_COMMAND = "arrange-_drawing";
    static final String MANUAL_MODE = "Manual mode";
    static final String MANUAL_MODE_COMMAND = "toggle-manual";
    static final String AUTO_ARRANGE = "Auto-Arrange";
    static final String AUTO_ARRANGE_COMMAND = "autoarrange";
    static final String SIMPLIFY_ALL_LINES = "Simplify all lines";
    static final String SIMPLIFY_ALL_LINES_COMMAND = "lines-simplify";
    static final String AUTO_SIMPLIFY = "Auto-Simplify";
    static final String AUTO_SIMPLIFY_COMMAND = "autosimplify";
    static final String RETRACE_ALL_LINES = "Retrace all lines";
    static final String RETRACE_ALL_LINES_COMMAND = "lines-retrace";
    static final String LOAD_DIAGRAM_BACKGROUND = "Load Background ...";
    static final String LOAD_DIAGRAM_BACKGROUND_COMMAND = "background-load";
    static final String TOGGLE_DIAGRAM_BACKGROUND_VISIBLE = "Background visible";
    static final String TOGGLE_DIAGRAM_BACKGROUND_VISIBLE_COMMAND = "background-toggle";
    static final String RECOGNIZE_DIAGRAM_BACKGROUND = "Recognize Background";
    static final String RECOGNIZE_DIAGRAM_BACKGROUND_COMMAND = "background-recognize";
    static final String RECOGNIZE_CURRENT_DIAGRAM = "Recognize Current";
    static final String RECOGNIZE_CURRENT_DIAGRAM_COMMAND = "background-setcurrentandrecognize";
    static final String DIAGRAM_LINE_TYPE_MENU = "Diagram line type";
    static final String DIAGRAM_LINE_TYPE_PREFIX = "_drawing-setlinetype";
}
}
