/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.WeakHashMap;

import lv.lumii.deek.data.DataFactory;
import lv.lumii.deek.data.DataProperties;
import lv.lumii.deek.style.Style;
import lv.lumii.deek.style.StyleFactory;
import lv.lumii.deek.ui.DDrawing;
import lv.lumii.deek.util.BlockSize;
import lv.lumii.deek.util.DLogger;
import lv.lumii.deek.util.DUtils;
import lv.lumii.layoutengine.*;
import lv.lumii.deek.util.Globals;
import lv.lumii.layoutengine.Container;
import lv.lumii.layoutengine.Element;
import org.w3c.dom.*;

import javax.swing.*;

/**
 * An abstract class serving as a base point for all elements in a diagram.
 */
public abstract class DElement {


public static final String NAME_PROPERTY_TAG = "name";
public static final String TEXT1_PROPERTY_TAG = "text1";
static final String LAYOUT_ID_ATTR = "layoutId";
// logger
protected static DLogger log = new DLogger();
private static Map<lv.lumii.layoutengine.Element, DElement> layoutMap = new WeakHashMap<>(256);
protected lv.lumii.layoutengine.Element layoutElement;
DataProperties data;
Style style;
private int m_nestingDepth = -1;
private boolean layouted;
private long uuid;


public DElement(lv.lumii.layoutengine.Element layoutElement) {
    linkToLayout(layoutElement);
    initElementDefaults();
}

DElement(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objectMap) {
    int layoutId = XMLHelper.loadInt(e, LAYOUT_ID_ATTR);
    lv.lumii.layoutengine.Element layoutElement = (Element) objectMap.get(layoutId);
    linkToLayout(layoutElement);

    data = DataFactory.createDataProperties(this);
    style = StyleFactory.createStyle(this);

    loadAttributesFromXML(e);
    createChildrenFromXML(e, objectMap);

}

public static void reset() {
    layoutMap.clear();
}

/**
 * Find the diagram for a given layout element.
 *
 * @param el layout element
 * @return diagram element
 */
static DElement getThis(lv.lumii.layoutengine.Element el) {
    return layoutMap.get(el);
}

static DLine getThisLine(Line el) {
    return (DLine) layoutMap.get(el);
}

static DBox getThisBox(lv.lumii.layoutengine.Box el) {
    return (DBox) layoutMap.get(el);
}

static DContainer getThisContainer(AbstractContainer el) {
    return (DContainer) layoutMap.get(el);
}

static DGroup getThisGroup(Container el) {
    return (DGroup) layoutMap.get(el);
}

//    public abstract DElement getHEFP(Point2D.Double p);

static DInsideLabel getThisInsideLabel(InsideLabel el) {
    return (DInsideLabel) layoutMap.get(el);
}

static DBoxLabel getThisBoxLabel(OutsideLabel.BoxOutsideLabel el) {
    return (DBoxLabel) layoutMap.get(el);
}

static DLineLabel getThisLineLabel(OutsideLabel.LineLabel el) {
    return (DLineLabel) layoutMap.get(el);
}

static void removeThis(lv.lumii.layoutengine.Element el) {
    layoutMap.remove(el);
}

/**
 * Returns the unique id of a diagram element
 *
 * @param el
 * @return
 */
public static long getThisId(lv.lumii.layoutengine.Element el) {
    return layoutMap.get(el).getUuid();
}

protected static void createChildrenFromXML(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objectMap) {

    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            String tag = c.getTagName();

            switch (tag) {
                case XMLHelper.BOXES_ELEMENT:
                    buildBoxes(c, objectMap);
                    break;

                case XMLHelper.LINES_ELEMENT:
                    buildLines(c, objectMap);
                    //
                    break;

                case XMLHelper.GROUPS_ELEMENT:
                    buildGroups(c, objectMap);
                    //
                    break;

                case XMLHelper.INSIDE_LABELS_ELEMENT:
                    buildInsideLabels(c, objectMap);
                    //
                    break;

                case XMLHelper.BOX_LABELS_ELEMENT:
                    buildBoxLabels(c, objectMap);
                    //
                    break;

                case XMLHelper.LINE_LABELS_ELEMENT:
                    buildLineLabels(c, objectMap);
                    //
                    break;

                default:
            }
        }
    }
}

static void buildBoxes(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            if (c.getTagName().equals(XMLHelper.BOX_ELEMENT)) {
                DBox x = new DBox(c, objMap);
            }
        }
    }
}

static void buildGroups(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            if (c.getTagName().equals(XMLHelper.GROUP_ELEMENT)) {
                DGroup x = new DGroup(c, objMap);
            }
        }
    }
}

static void buildLines(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            if (c.getTagName().equals(XMLHelper.LINE_ELEMENT)) {
                DLine x = new DLine(c, objMap);
            }
        }
    }
}

static void buildLineLabels(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            if (c.getTagName().equals(XMLHelper.LINE_LABEL_ELEMENT)) {
                DLineLabel x = new DLineLabel(c, objMap);
            }
        }
    }
}

static void buildBoxLabels(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            if (c.getTagName().equals(XMLHelper.BOX_LABEL_ELEMENT)) {
                DBoxLabel x = new DBoxLabel(c, objMap);
            }
        }
    }
}

static void buildInsideLabels(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
    NodeList children = e.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child instanceof org.w3c.dom.Element) {
            org.w3c.dom.Element c = (org.w3c.dom.Element) child;
            if (c.getTagName().equals(XMLHelper.INSIDE_LABEL_ELEMENT)) {
                DInsideLabel x = new DInsideLabel(c, objMap);
            }
        }
    }
}

private void initElementDefaults() {
    uuid = Globals.nextUuid();

    data = DataFactory.createDataProperties(this);
    style = StyleFactory.createStyle(this);

    style.delayUpdate(true);
    setupElementStyle(style);
    style.delayUpdate(false);

    setName(getDefaultName());
    setText1(DUtils.getRandomPhrase());
}

/**
 * Returns the diagram this element belongs to.
 *
 * @return Diagram instance
 */
public DDiagram getDiagram() {
    DDiagram d = (DDiagram) DElement.getThis(getLayoutElement().getDiagram());
    if (d == null)
        System.out.println(String.format("Uzmanību – %s.getDiagram()==null", this.getClass().getSimpleName()));
    return d;
}

public void setDiagram(DDiagram diagram) {
    throw new UnsupportedOperationException("Element moving between diagrams is not supported yet");
}

/**
 * Returns the drawing this element belongs to.
 *
 * @return drawing
 */
public DDrawing getDrawing() {
    return getDiagram().getDrawing();
}

/**
 * Returns the owner of this element.
 *
 * @return owner of the element
 */
public abstract DElement getOwner();

/**
 * Removes all child elements, then removes this element from the diagram.
 */
public final void remove() {
    removeChildren();
    removeThis();
    updateAll();
}

/**
 * Remove all child elements for this element.
 */
protected abstract void removeChildren();

protected void removeThis() {
    if (layoutElement != null) {
        layoutMap.remove(layoutElement);
        layoutElement.remove();
    }
}

/**
 * Draws this element on the diagram canvas.
 *
 * @param g Graphics context
 */
public final void draw(Graphics2D g) {

    drawSelf(g);

    drawDecorations(g);

    if (isHover())
        drawTestInfo(g);

    drawChildren(g);
}

protected void drawSelf(Graphics2D g) {
}

protected void drawChildren(Graphics2D g) {
}

protected void drawDecorations(Graphics2D g) {
}

//	// --------- to do ----------------
protected boolean isHover() {
    // TODO
    return false;
}

protected boolean isCurrent() {
    return getDrawing().getCurrent() == this;
}

/**
 * Checks whether the element or one of its subelements is hit by the point
 *
 * @param p Point to be checked against.
 * @return true if hit
 */
public boolean isHit(Point2D.Double p) {
    if (isSelfHit(p))
        return true;
    return isChildHit(p);
}

/**
 * Checks whether the element is hit by the point
 *
 * @param p Point to be checked against.
 * @return true if hit
 */
public abstract boolean isSelfHit(Point2D.Double p);

/**
 * Checks whether one of the child elements is hit by the point
 *
 * @param p point to be checked against
 * @return true if any child is hit
 */
public abstract boolean isChildHit(Point2D.Double p);

/**
 * Tests this element and the child elements against the point.
 *
 * @param p point to be checked against
 * @return Child element hit by the point; this element, if it is hit; otherwise null
 */
public DElement getHit(Point2D.Double p) {
    DElement e = getChildHit(p);
    if (e != null)
        return e;
    return getSelfHit(p);
}

/**
 * Tests this element against the point.
 *
 * @param p point to be checked against
 * @return this element if it is hit by the point, otherwise null
 */
public DElement getSelfHit(Point2D.Double p) {
    if (isSelfHit(p))
        return this;
    return null;
}

/**
 * Tests child elements against the point
 *
 * @param p point to be checked against
 * @return Child element hit by the point; otherwise null.
 */
public abstract DElement getChildHit(Point2D.Double p);

/**
 * Finds rectangular bounds containing this element
 *
 * @return Rectangle
 */
public Rectangle2D.Double getBounds() {
    return null;
}

/**
 * Returns the instance of the global settings object.
 *
 * @return the instance of the global settings object.
 */
public Globals getGlobals() {
    return Globals.getInstance();
}

public int getNestingDepth() {
    if (m_nestingDepth >= 0) {
        return m_nestingDepth;
    }

    if (getOwner() == null) {
        m_nestingDepth = 0;
    } else {
        m_nestingDepth = getOwner().getNestingDepth() + 1;
    }
    return m_nestingDepth;
}

protected void invalidateStructure() {
    m_nestingDepth = -1;
}

protected void invalidateGeometry() {
}

protected void linkToLayout(lv.lumii.layoutengine.Element layoutElement) {
    this.layoutElement = layoutElement;
    layoutMap.put(layoutElement, this);
    this.layouted = true;
}

protected void unlinkFromLayout() {
    if (layoutElement != null) {
        layoutMap.remove(layoutElement);
        layoutElement.remove(); // TODO ??? vai to vajag?
        layoutElement = null;
        this.layouted = false;
    }
}

boolean isLayouted() {
    return this.layouted;
}

void setLayouted(boolean hasLayout) {
    throw new UnsupportedOperationException("layout activating/deactivating not supported yet");
}

/**
 * Find the layout element for this diagram element.
 *
 * @return layout element
 */
abstract lv.lumii.layoutengine.Element getLayoutElement();

//public final double getSpacing() {
//    return style.getSpacing();
//}

//public abstract void setSpacing(double newSpacing);

/**
 * Returns the unique identifier for this element
 *
 * @return the unique identifier for this element
 */
public long getUuid() {
    return uuid;
}

/**
 * Calculates the default name for a diagram element.
 * Will be used when no name is explicitly set
 *
 * @return the default name
 */
protected String getDefaultName() {
//    return this.getClass().getSimpleName() + Long.toString(uuid);
    String prefix = "";
    if (this instanceof DBox)
        prefix = "Kaste ";
    else if (this instanceof DLine)
        prefix = "Līnija ";
    else if (this instanceof DLabel)
        prefix = "Birka ";
    else if (this instanceof DGroup)
        prefix = "Grupa ";

    return prefix + Long.toString(uuid);
}

/**
 * Returns the name for this element
 *
 * @return name of this element
 */
public String getName() {
    if (data == null)
        return getDefaultName();

    String name = data.getDataString(NAME_PROPERTY_TAG);
    if (name == null)
        return getDefaultName();

    return name;
}

/**
 * Set the name for this element.
 *
 * @param name new name
 */
public void setName(String name) {
    data.setDataString(NAME_PROPERTY_TAG, name);
    recalculate();
}

public String getText1() {
    return data.getDataString(TEXT1_PROPERTY_TAG);
}

public void setText1(String text1) {
    data.setDataString(TEXT1_PROPERTY_TAG, text1);
}

/**
 * Returns the Style object for this element
 *
 * @return the Style object for this element
 */
public Style getStyle() {
    return style;
}

/**
 * Defines the initial style for this diagram element
 *
 * @param style
 */
protected abstract void setupElementStyle(Style style);

/**
 * Return the spacing distance for this element.
 *
 * @return the spacing
 */
public double getSpacing() {
    return getLayoutElement().getSpacing();
}

/**
 * Set the spacing distance for this element.
 *
 * @param newSpacing spacing
 */
public void setSpacing(double newSpacing) {
    getLayoutElement().setSpacing(newSpacing);
}

BlockSize measure() {
    return new BlockSize(0, 0);
}

/**
 * Sets this element's style as a <strong>copy</strong> of source's style
 *
 * @param source
 */
void cloneStyleFrom(DElement source) {
    this.style = source.style;
}

/**
 * Sets this element's style to the <strong>same</strong> style instance as the source
 * (i.e., changes in the style will affect all elements sharing this style. Use with care.)
 *
 * @param source
 */
void setStyleFrom(DElement source) {
    this.style = source.style;
}

/**
 * Gets the Paint which will be used for the inner area of this element
 *
 * @return the paint
 */
public Paint getFillPaint() {
    Rectangle2D.Double r = getBounds();
    if (r != null)
        return style.getPaint(r);
    else
        return style.getPaint();
}

/**
 * @param c
 */
public void setFillColor(Color c) {
    style.setFillColor(c);
}

protected String getTestInfoString() {
    return String.format("#%d: ", getUuid());
}

protected void drawTestInfo(Graphics2D g) {
}

protected JMenu getStyleSubmenu() {
    return style.getMenu(Style.StyleMenuType.ELEMENT);
}

/**
 * Builds and returns the context menu for this element.
 *
 * @return the context menu
 */
public JPopupMenu getElementMenu() {
    JPopupMenu menu = new JPopupMenu();

    menu.add(menuCommand(Commands.DELETE_ELEMENT, Commands.DELETE_ELEMENT_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            remove();
            updateAll();
        }
    }));

    menu.addSeparator();

    // content
    menu.add(menuCommand(Commands.EDIT_ELEMENT_CONTENT, Commands.EDIT_ELEMENT_CONTENT_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {

            String s = (String) JOptionPane.showInputDialog(
//                        _frame,
                    null,
                    "New name:",
                    "Element name",
                    JOptionPane.QUESTION_MESSAGE,
                    null,
                    null,
                    getName());
            if (s == null) {
                return;
            }

            setName(s);
            recalculate();
            setCurrent();
            updateAll();

        }
    }));

    menu.add(menuCommand(Commands.EDIT_ELEMENT_LAYOUT, Commands.EDIT_ELEMENT_LAYOUT_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            double spacing = getSpacing();
            double newSpacing;
            String s = JOptionPane.showInputDialog(
//                        _frame,
                    null,
                    "Enter the new spacing:",
                    spacing);

            if (s != null) {

                try {
                    newSpacing = Double.parseDouble(s);
                } catch (NumberFormatException ex) {
                    newSpacing = getGlobals().getCurrentBoxSpacing();
                }

                newSpacing = Math.max(newSpacing, Globals.spacingMin);

                setSpacing(newSpacing);
                setCurrent();
                updateAll();

            }
        }
    }));

    menu.add(getStyleSubmenu());

    return menu;
}

protected JMenu getLineTypeSubmenu(String title, String prefix, DLine.DLineType currentLineType, ActionListener listener) {

    JMenuItem item;
    JMenu lineTypeSubmenu = new JMenu(title);

    for (DLine.DLineType lt : DLine.DLineType.values()) {
        item = new JCheckBoxMenuItem(lt.toString());
        if (currentLineType == lt)
            item.setSelected(true);
        item.addActionListener(listener);
        item.setActionCommand(lt.toString());
        lineTypeSubmenu.add(item);
    }
    return lineTypeSubmenu;
}

public abstract void recalculate();

/**
 * Sets this elements row in the parents grid, if any
 *
 * @param row grid row
 */
public void setGridRow(Integer row) {
    if (getOwner() == null)
        return;
    DElement e = getOwner();
    if (!(e instanceof DContainer))
        return;
    DContainer host = (DContainer) e;
    if (host.getLayoutConstraints().getType() != LayoutConstraints.ConstraintType.GRID)
        return;
    LayoutConstraints.GridLayoutConstraints glc = (LayoutConstraints.GridLayoutConstraints) host.getLayoutConstraints();

    glc.setRow(getLayoutElement(), row);
}

/**
 * Sets this elements column in the parents grid, if any
 *
 * @param column grid column
 */
public void setGridColumn(Integer column) {
    if (getOwner() == null)
        return;
    DElement e = getOwner();
    if (!(e instanceof DContainer))
        return;
    DContainer host = (DContainer) e;
    if (host.getLayoutConstraints().getType() != LayoutConstraints.ConstraintType.GRID)
        return;
    LayoutConstraints.GridLayoutConstraints glc = (LayoutConstraints.GridLayoutConstraints) host.getLayoutConstraints();

    glc.setColumn(getLayoutElement(), column);
}

protected JMenuItem menuCommand(String text, String command, ActionListener listener) {
    JMenuItem item = new JMenuItem(text);
    item.setActionCommand(command);
    item.addActionListener(listener);
    return item;
}

protected JMenuItem menuCommand(String text, String command, ActionListener listener, boolean selected) {
    JMenuItem item = new JCheckBoxMenuItem(text);
    item.setActionCommand(command);
    item.addActionListener(listener);
    item.setSelected(selected);
    return item;
}

protected void setCurrent() {
    getDiagram().getDrawing().setCurrent(this);
}

public void updateAll() {

    if (getDiagram() == null) {
        return;
    }

    DDrawing d = getDiagram().getDrawing();
    if (d == null)
        return;

    d.requestUpdate();
}

public void updateThis() {
    recalculate();
}

protected void startTransaction() {
    getDiagram().startTransaction();
}

protected void endTransaction() {
    getDiagram().endTransaction();
}

org.w3c.dom.Element saveToXML(org.w3c.dom.Document doc, LinkedHashMap<Object, Integer> idMap) {
    String tagName;
    String objType = this.getClass().getSimpleName();
    switch (objType) {
        case "DDiagram":
            tagName = XMLHelper.DIAGRAM_ELEMENT;
            break;

        case "DBox":
            tagName = XMLHelper.BOX_ELEMENT;
            break;

        case "DGroup":
            tagName = XMLHelper.GROUP_ELEMENT;
            break;

        case "DLine":
            tagName = XMLHelper.LINE_ELEMENT;
            break;

        case "DLineLabel":
            tagName = XMLHelper.LINE_LABEL_ELEMENT;
            break;

        case "DBoxLabel":
            tagName = XMLHelper.BOX_LABEL_ELEMENT;
            break;

        case "DInsideLabel":
            tagName = XMLHelper.INSIDE_LABEL_ELEMENT;
            break;


        default:
            System.out.println("Unprocessed type @ save to xml: " + objType);
            return null;
    }
    org.w3c.dom.Element e = doc.createElement(tagName);

    Integer idLayout = idMap.get(getLayoutElement());
    e.setAttribute(LAYOUT_ID_ATTR, idLayout.toString());

    saveAttributesToXML(e);
    saveChildrenToXML(e, doc, idMap);

    return e;
}

void saveAttributesToXML(org.w3c.dom.Element e) {
    e.setAttribute("uuid", Long.toString(uuid));

//    org.w3c.dom.Element eStyle = e.getOwnerDocument().createElement(XMLHelper.STYLE_ELEMENT);
//    org.w3c.dom.Element eData = e.getOwnerDocument().createElement(XMLHelper.DATA_ELEMENT);
    e.appendChild(style.saveToXML(e.getOwnerDocument()));
    e.appendChild(data.saveToXML(e.getOwnerDocument()));
}

void loadAttributesFromXML(org.w3c.dom.Element e) {
    Node node;

    uuid = XMLHelper.loadLong(e, "uuid");

    node = e.getElementsByTagName(XMLHelper.STYLE_ELEMENT).item(0);
    if (node != null) {
        style = StyleFactory.createStyleFromXML(this, (org.w3c.dom.Element) node);
    }

    node = e.getElementsByTagName(XMLHelper.DATA_ELEMENT).item(0);
    if (node != null) {
        data = DataFactory.createDataPropertiesFromXML(this, (org.w3c.dom.Element) node);
    }

}

protected void saveChildrenToXML(org.w3c.dom.Element e, org.w3c.dom.Document doc, LinkedHashMap<Object, Integer> idMap) {
}

private static class Commands {
    // do not use ':' in commands and prefixes, as it is used as separator here

    static final String DELETE_ELEMENT = "Delete";
    static final String DELETE_ELEMENT_COMMAND = "element-delete";
    static final String EDIT_ELEMENT_CONTENT = "Edit ...";
    static final String EDIT_ELEMENT_CONTENT_COMMAND = "element-editdialog";

    //    static final String EDIT_ELEMENT_FILLCOLOR = "Set fill color";
//    static final String EDIT_ELEMENT_FILLCOLOR_COMMAND = "element-set-fill-color";
//
//    static final String EDIT_ELEMENT_BORDERCOLOR = "Set border color";
//    static final String EDIT_ELEMENT_BORDERCOLOR_COMMAND = "element-set-border-color";
//
//    static final String EDIT_ELEMENT_STYLE = "Set color";
//    static final String EDIT_ELEMENT_STYLE_COMMAND = "element-set-color";
    static final String EDIT_ELEMENT_LAYOUT = "Layout ...";
    static final String EDIT_ELEMENT_LAYOUT_COMMAND = "element-layoutdialog";

//    static final String DUMP_ELEMENT = "Dump";
//    static final String DUMP_ELEMENT_COMMAND = "element-dump";
//
//    static final String SIMPLIFY_LINE = "Simplify";
//    static final String SIMPLIFY_LINE_COMMAND = "line-simplify";
//
//    static final String RETRACE_LINE = "Retrace";
//    static final String RETRACE_LINE_COMMAND = "line-retrace";
//
//
//    static final String ARRANGE_STYLE_MENU = "Arrange Style";
//    static final String ARRANGE_STYLE_PREFIX = "arrangestyle";
//
//    static final String ARRANGE_STYLE_FLOW_DIRECTION_MENU = "Flow Direction";
//    static final String ARRANGE_STYLE_FLOW_DIRECTION_PREFIX = "flowdirection";
//
//    static final String CONSTRAINT_TYPE_MENU = "Constraint Type";
//    static final String CONSTRAINT_TYPE_PREFIX = "constrainttype";

//		static final String GRID_SET_DIMENSIONS = "Set grid dimensions...";
//		static final String GRID_SET_DIMENSIONS_COMMAND = "constraints-grid-setdimensions";

//    static final String GRID_SET_ROWCOUNT_MENU = "Row count";
//    static final String GRID_SET_ROWCOUNT_PREFIX = "grid-rowcount";
//
//    static final String GRID_SET_COLCOUNT_MENU = "Column count";
//    static final String GRID_SET_COLCOUNT_PREFIX = "grid-colcount";
//
//    static final int GRID_MAXROWS = 5;
//    static final int GRID_MAXCOLS = 7;
//
//    static final String GRID_SET_ROW_MENU = "Set row";
//    static final String GRID_SET_ROW_PREFIX = "grid-row";
//
//    static final String GRID_SET_COL_MENU = "Set column";
//    static final String GRID_SET_COL_PREFIX = "grid-col";
//
//    static final String GRID_ROWCOL_NONE = "none";
//    static final String GRID_ROWCOL_NONE_STRINGVALUE = "none";
//    static final int GRID_ROWCOL_NONE_INTVALUE = -1;
//
//    static final String GRIDIFY_CHILDREN = "Gridify children";
//    static final String GRIDIFY_CHILDREN_COMMAND = "constraints-grid-gridifychildren";

//    static final String LABELIFY_CHILDREN = "Child boxes to labels";
//    static final String LABELIFY_CHILDREN_COMMAND = "test-labelifychildren";
//
//    static final String CONVERT_TO_BOX = "Convert to box";
//    static final String CONVERT_TO_BOX_COMMAND = "group-converttobox";
//
//    static final String CONVERT_TO_GROUP = "Convert to group";
//    static final String CONVERT_TO_GROUP_COMMAND = "group-converttogroup";
//
//    static final String ARRANGE_DIAGRAM = "Arrange Diagram";
//    static final String ARRANGE_DIAGRAM_COMMAND = "arrange-_drawing";
//
//    static final String MANUAL_MODE = "Manual mode";
//    static final String MANUAL_MODE_COMMAND = "toggle-manual";
//
//    static final String AUTO_ARRANGE = "Auto-Arrange";
//    static final String AUTO_ARRANGE_COMMAND = "autoarrange";
//
//    static final String SIMPLIFY_ALL_LINES = "Simplify all lines";
//    static final String SIMPLIFY_ALL_LINES_COMMAND = "lines-simplify";
//
//    static final String AUTO_SIMPLIFY = "Auto-Simplify";
//    static final String AUTO_SIMPLIFY_COMMAND = "autosimplify";
//
//    static final String RETRACE_ALL_LINES = "Retrace all lines";
//    static final String RETRACE_ALL_LINES_COMMAND = "lines-retrace";
//
//    static final String LOAD_DIAGRAM_BACKGROUND = "Load Background ...";
//    static final String LOAD_DIAGRAM_BACKGROUND_COMMAND = "background-load";
//
//    static final String TOGGLE_DIAGRAM_BACKGROUND_VISIBLE = "Background visible";
//    static final String TOGGLE_DIAGRAM_BACKGROUND_VISIBLE_COMMAND = "background-toggle";
//
//    static final String RECOGNIZE_DIAGRAM_BACKGROUND = "Recognize Background";
//    static final String RECOGNIZE_DIAGRAM_BACKGROUND_COMMAND = "background-recognize";
//
//    static final String RECOGNIZE_CURRENT_DIAGRAM = "Recognize Current";
//    static final String RECOGNIZE_CURRENT_DIAGRAM_COMMAND = "background-setcurrentandrecognize";
//
//    static final String CLEAR_DIAGRAM = "Clear Diagram";
//    static final String CLEAR_DIAGRAM_COMMAND = "clear-_drawing";
//
//    static final String DIAGRAM_LINE_TYPE_MENU = "Diagram line type";
//    static final String DIAGRAM_LINE_TYPE_PREFIX = "_drawing-setlinetype";
//    static final String LINE_TYPE_MENU = "Line Type";
//    static final String LINE_TYPE_PREFIX = "line-settype";
//
//    static final String LINE_START_SIDES_MENU = "Start sides";
//    static final String LINE_START_SIDES_PREFIX = "line-setstartsides";
//    static final String LINE_END_SIDES_MENU = "End sides";
//    static final String LINE_END_SIDES_PREFIX = "line-setendsides";
//
//    static final String BOX_INCOMING_LINE_SIDES_MENU = "Incoming line sides";
//    static final String BOX_INCOMING_LINE_SIDES_PREFIX = "boxlines-setincomingsides";
//    static final String BOX_OUTGOING_LINE_SIDES_MENU = "Outgoing line sides";
//    static final String BOX_OUTGOING_LINE_SIDES_PREFIX = "boxlines-setoutgoingsides";
//    static final String BOX_RESET_LINE_SIDES = "Reset sides";
//    static final String BOX_RESET_LINE_SIDES_COMMAND = "RESET";
//
//    static final String LINE_LABEL_ORIENTATION_MENU = "Position";
//    static final String LINE_LABEL_ORIENTATION_PREFIX = "linelabel-setorientation";
//
//    static final String OUTSIDE_LABEL_AUTOFLIP = "AutoFlip";
//    static final String OUTSIDE_LABEL_AUTOFLIP_COMMAND = "outsidelabel-setautoflip";
//
//    static final String LINE_LABEL_ROTATE = "Rotate";
//    static final String LINE_LABEL_ROTATE_COMMAND = "outsidelabel-setrotate";
//
//    static final String OUTSIDE_LABEL_FLIP = "Flip";
//    static final String OUTSIDE_LABEL_FLIP_COMMAND = "outsidelabel-flip";
}


}
