/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import lv.lumii.deek.style.Style;
import lv.lumii.deek.util.BlockSize;
import lv.lumii.deek.util.DUtils;
import lv.lumii.deek.util.Globals;
import lv.lumii.layoutengine.Label;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;

/**
 * Abstract class describing labels.
 * <p/>
 * Labels are simpler rectangular elements of the diagram, which cannot have lines connected to them.
 * <p/>
 * There are several label subclasses for different roles.
 */
public abstract class DLabel extends DElement {

private Pack contentPack = null;

public DLabel(Label label) {
    super(label);
    initDefaults();
}

DLabel(org.w3c.dom.Element element, LinkedHashMap<Integer, Object> objMap) {
    super(element, objMap);
}

private void initDefaults() {
    // for test:
//    if (Globals.randomGenerator.nextBoolean()) {
//        addText(getName());
//        addSeparator();
//        addText("aaa");
//    }
}

@Override
public lv.lumii.layoutengine.Label getLayoutElement() {
    throw new UnsupportedOperationException("Label is an abstract creature and should not have instances");
}

@Override
public void removeChildren() {
    // no children yet
}

public DLabel addText(String text) {
    if (contentPack == null)
        contentPack = new Pack(this);
    contentPack.addText(text);
    recalculate();
    return this;
}

public DLabel addSeparator() {
    if (contentPack == null)
        contentPack = new Pack(this);
    contentPack.addSeparator();
    recalculate();
    return this;
}

@Override
protected void setupElementStyle(Style style) {

    style.delayUpdate(true);
    style.setFillColor(Globals.LABEL_COLOR);
    style.setStrokeColor(Globals.LABEL_BORDER_COLOR);
    style.setFontSize(12);
    style.setTextColor(new Color(0, 64, 0));
    style.setFontBold(true);
    style.setPadding(0);
    style.delayUpdate(false);

}

@Override
protected JMenu getStyleSubmenu() {
    return style.getMenu(Style.StyleMenuType.BOX);
}

@Override
protected void drawSelf(Graphics2D g) {
    // TODO
    Shape thisShape = getShape();
    Rectangle2D.Double shapeBounds = (Rectangle2D.Double) thisShape.getBounds2D();

    g.setPaint(style.getPaint(shapeBounds));
    g.fill(thisShape);

    if (isCurrent())
        g.setColor(Globals.CURRENT_COLOR);
    else
        g.setColor(style.getStrokeColor());
    g.draw(thisShape);

    drawContent(g);
    System.out.println("Label.drawSelf");
}

@Override
protected void drawDecorations(Graphics2D g) {

    if (Globals.isDeveloperMode()) {
    }
    // for testing
    g.setColor(Globals.LABEL_BORDER_COLOR_PHANTOM);
    g.setStroke(Globals.GRID_BORDER_STROKE);

    g.draw(getBounds());


//    super.drawDecorations(g);
}

@Override
BlockSize measure() {

    if (contentPack != null) {
        return contentPack.measure().pad(style.getPadding());
    } else {
        BlockSize dim = DUtils.measureText2(style.getFont(), getName());
        return dim.pad(style.getPadding());
    }
}

/**
 * Is this label flipped?
 *
 * @return true if flipped
 */
abstract boolean isFlipped();

protected void drawContent(Graphics2D g) {

    Rectangle2D.Double r = getBounds();

    if (contentPack != null) {

        if (isFlipped()) {
            contentPack.draw(g, style.getPadding(), style.getPadding(), r.height - 2 * style.getPadding(), r.width - 2 * style.getPadding());
        } else {
            contentPack.draw(g, style.getPadding(), style.getPadding(), r.width - 2 * style.getPadding(), r.height - 2 * style.getPadding());
        }
    } else {
        g.setFont(style.getFont());
        g.setColor(style.getTextColor());

        BlockSize dim = DUtils.measureText(g, getName());

        float xPos = (float) (r.getCenterX() - dim.width / 2 + style.getPadding() - r.x);
        float yPos = (float) (r.getCenterY() + dim.height / 4 + style.getPadding() - r.y);

        g.drawString(getName(), 0, (float) (dim.height * 0.75));
    }
}

private Shape getShape() {
    return getBounds();
}

Point2D.Double getCenter() {
    return new Point2D.Double(getBounds().getCenterX(), getBounds().getCenterY());
}

@Override
public boolean isChildHit(Point2D.Double p) {
    return false;
}

@Override
public boolean isSelfHit(Point2D.Double p) {
    return getLayoutElement().getBounds().contains(p);
}

@Override
public DElement getChildHit(Point2D.Double p) {
    return null;
}

@Override
public DElement getOwner() {
    throw new UnsupportedOperationException("Label is an abstract creature");
}

@Override
public Rectangle2D.Double getBounds() {
    return getLayoutElement().getBounds();
}

@Override
public JPopupMenu getElementMenu() {
    JPopupMenu menu = super.getElementMenu();

//        menu.addSeparator();

    // ...

    return menu;
}

@Override
public void recalculate() {
    BlockSize dim = measure();
    ensureSize(dim);
}

protected void ensureSize(BlockSize dim) {
    setMinSize(dim.width, dim.height);
}

public void setMinSize(double minWidth, double minHeight) {
    if (!isFlipped())
        getLayoutElement().setMinSize(minWidth, minHeight);
    else
        getLayoutElement().setMinSize(minHeight, minWidth);
}

public abstract void resize(Rectangle2D.Double r);

//    public abstract void resize(double width, double height);


}
