/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;

import lv.lumii.deek.style.BoxShapeType;
import lv.lumii.deek.style.Style;
import lv.lumii.deek.util.HitTest;
import lv.lumii.layoutengine.Box;
import lv.lumii.layoutengine.LayoutConstraints;
import lv.lumii.layoutengine.Line;
import lv.lumii.layoutengine.OutsideLabel.LineLabel;
import lv.lumii.deek.util.Globals;
import lv.lumii.deek.util.Drawer;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.swing.*;

/**
 * Line is a one-dimensional diagram element, connecting two boxes with each other.
 * <p/>
 * There are several geometric types of lines: simple (straight) lines, poly lines and ortho lines (lines,
 * which consist of orthogonal segments, parallel to one of the coordinate axis).
 */
public class DLine extends DElement {


private static Color DEFAULT_LINE_COLOR = new Color(0, 64, 64);
private DLineType _lineType = DLineType.ORTHO;

public DLine(lv.lumii.layoutengine.Line line, DLineType lineType) {
    super(line);
    _lineType = lineType;
    if (!Globals.isDeveloperMode()) {

        startTransaction();

        addTextAtStart("0", LineLabel.Orientation.CLOCKWISE);
        addTextAtStart("contains", LineLabel.Orientation.COUNTERCLOCKWISE);
        addTextAtEnd("*", LineLabel.Orientation.COUNTERCLOCKWISE);
        addTextAtEnd("belongs to", LineLabel.Orientation.CLOCKWISE);
        addTextAtMiddle(getName());

        endTransaction();
    }
}

DLine(org.w3c.dom.Element c, LinkedHashMap<Integer, Object> objMap) {
    super(c, objMap);
}

@Override
protected void setupElementStyle(Style style) {
    style.setStrokeColor(DEFAULT_LINE_COLOR);
}

@Override
protected JMenu getStyleSubmenu() {
    return style.getMenu(Style.StyleMenuType.LINE);
}

protected void adjustLinePoints(ArrayList<Point2D.Double> points) {
    if (points == null)
        return;

    if (points.size() < 2)
        return;

    // adjust start point
    BoxShapeType startShape = getStart().style.getBoxShapeType();
    // pielāgot nogriezni points(0)..points(1)

    // adjust end point
    BoxShapeType endShape = getEnd().style.getBoxShapeType();
    // pielāgot nogriezni points(size-1)..points(size-2)

}

/**
 * Creates a new label for this line.
 * The position for the new label is determined by the given point.
 * If this point does not lie on the line then it will be shifted to the line.
 *
 * @param width          width of the label
 * @param height         height of the label
 * @param insertPoint    point which determines the label position
 * @param orientation    how to place the new label relative to the line trace
 * @param constraintType layout constraint type for the new label
 * @param lineSpacing    spacing for the new label
 * @return the label created
 */
public DLineLabel createLabel(double width, double height, Point2D.Double insertPoint, LineLabel.Orientation orientation, LayoutConstraints.ConstraintType constraintType, double lineSpacing) {
    return new DLineLabel(getLayoutElement().createLabel(width, height, insertPoint, orientation, constraintType, lineSpacing));
}

/**
 * Creates a new label for this line.
 *
 * @param width       width of the label
 * @param height      height of the label
 * @param insertPoint insertPoint point which determines the label position
 * @param orientation how to place the new label relative to the line trace
 * @return the label created
 */
public DLineLabel createLabel(double width, double height, Point2D.Double insertPoint, LineLabel.Orientation orientation) {
    return new DLineLabel(getLayoutElement().createLabel(width, height, insertPoint, orientation, LayoutConstraints.ConstraintType.NONE, Globals.getCurrentLineSpacing()));
}

/**
 * Creates a new label for this line.
 *
 * @param width          width of the label
 * @param height         height of the label
 * @param position       position of the label along the line trace (0 denotes line start, 1 denotes line end)
 * @param orientation    how to place the new label relative to the line trace
 * @param constraintType layout constraint type for the new label
 * @param lineSpacing    spacing for the new label
 * @return the label created
 */
public DLineLabel createLabel(double width, double height, double position, LineLabel.Orientation orientation, LayoutConstraints.ConstraintType constraintType, double lineSpacing) {
    return new DLineLabel(getLayoutElement().createLabel(width, height, position, orientation, constraintType, lineSpacing));
}

/**
 * Creates a new label for this line.
 *
 * @param width       width of the label
 * @param height      height of the label
 * @param position    position of the label along the line trace (0 denotes line start, 1 denotes line end)
 * @param orientation how to place the new label relative to the line trace
 * @return the label created
 */
public DLineLabel createLabel(double width, double height, double position, LineLabel.Orientation orientation) {
    return new DLineLabel(getLayoutElement().createLabel(width, height, position, orientation, LayoutConstraints.ConstraintType.NONE, Globals.getCurrentLineSpacing()));
}

private lv.lumii.layoutengine.Line.LineType getInternalLineType() {
    return getLayoutElement().getType();
}

/**
 * Returns the type of this line
 *
 * @return line type
 */
public DLineType getLineType() {
    return _lineType;
}

/**
 * Changes the type for this line.
 * The previous geometry is preserved as much as possible.
 *
 * @param lineType the new type for this line
 */
public void setLineType(DLineType lineType) {
    _lineType = lineType;
    getLayoutElement().setType(_lineType.getInternalType());
}

public DLineType getBaseLineType() {
    if (_lineType == DLineType.ORTHO_SPLINED)
        return DLineType.ORTHO;
    else if (_lineType == DLineType.POLY_SPLINED)
        return DLineType.POLY;
    else
        return _lineType;
}

@Override
public lv.lumii.layoutengine.Line getLayoutElement() {
    return (lv.lumii.layoutengine.Line) layoutElement;
}

@Override
public DElement getOwner() {
    return getThisContainer(getLayoutElement().getOwner());
}

/**
 * Return the corner points for this line (2 for a straight line, 2 or more for poly and ortho lines)
 *
 * @return list of the points
 */
public ArrayList<Point2D.Double> getPoints() {
    return getLayoutElement().getPoints();
}

/**
 * Sets the line geometry to the new points.
 * The points may be corrected to match the line type. The whole
 * diagram is reordered afterwards, or at the end of the current transaction.
 *
 * @param points the new corner points; their exact meaning may vary depending on the line type
 */
public void setPoints(ArrayList<Point2D.Double> points) {
    getLayoutElement().setPoints(points);
}

/**
 * Sets the line geometry to the new points.
 * The points may be corrected to match the line type. The whole
 * diagram is reordered afterwards, or at the end of the current transaction.
 *
 * @param points      the new corner points; their exact meaning may vary depending on the line type
 * @param autoCleanup whether to clean up the resulting line in order to minimize corner count
 */
public void setPoints(ArrayList<Point2D.Double> points, boolean autoCleanup) {
    getLayoutElement().setPoints(points, autoCleanup);
}

/**
 * Sets the line geometry to the new points.
 * The points may be corrected to match the line type. The whole
 * diagram is reordered afterwards, or at the end of the current transaction.
 *
 * @param points   the new corner points; their exact meaning may vary depending on the line type
 * @param newStart the new start element for this line
 * @param newEnd   the new end element for this line
 */
public void setPoints(ArrayList<Point2D.Double> points, DBox newStart, DBox newEnd) {
    getLayoutElement().setPoints(points, newStart.getLayoutElement(), newEnd.getLayoutElement());
}

/**
 * Sets the line geometry to the new points.
 * The points may be corrected to match the line type. The whole
 * diagram is reordered afterwards, or at the end of the current transaction.
 *
 * @param points      the new corner points; their exact meaning may vary depending on the line type
 * @param newStart    the new start element for this line
 * @param newEnd      the new end element for this line
 * @param autoCleanup whether to clean up the resulting line in order to minimize corner count
 */
public void setPoints(ArrayList<Point2D.Double> points, DBox newStart, DBox newEnd, boolean autoCleanup) {
    getLayoutElement().setPoints(points, newStart.getLayoutElement(), newEnd.getLayoutElement(), autoCleanup);
}

/**
 * Returns the starting point for this line.
 *
 * @return the starting point for this line.
 */
public Point2D.Double getStartPoint() {
    return getLayoutElement().getStartPoint();
}

/**
 * Returns the ending point for this line.
 *
 * @return the ending point for this line.
 */
public Point2D.Double getEndPoint() {
    return getLayoutElement().getEndPoint();
}

/**
 * Return the start element for this line
 *
 * @return Box at the line start
 */
public final DBox getStart() {
    return DBox.getThisBox((Box) getLayoutElement().getStart());
}

/**
 * Return the end element for this line
 *
 * @return Box at the line end
 */
public final DBox getEnd() {
    return DBox.getThisBox((Box) getLayoutElement().getEnd());
}

@Override
protected void drawSelf(Graphics2D g) {

    if (isHover()) {
        g.setColor(Globals.HOVER_COLOR);
    } else if (isCurrent()) {
        g.setColor(Globals.CURRENT_COLOR);
    } else {
        g.setColor(style.getStrokeColor());
    }

    switch (getLineType()) {
        case ORTHO:
            g.setStroke(style.getStroke());
            Drawer.drawDecoratedOrthoLine(g, getPoints(),
                    style.getLineStartDecoration(),
                    style.getLineEndDecoration());
            break;

        case ORTHO_SPLINED:
            g.setStroke(style.getStroke());
            Drawer.drawDecoratedSpline(g, getPoints(),
                    style.getLineStartDecoration(),
                    style.getLineEndDecoration());
            g.setStroke(Globals.LINE_STROKE_PHANTOM);
            g.setColor(Globals.LINE_COLOR_PANTOM);
            Drawer.drawDecoratedOrthoLine(g, getPoints());
            break;

        case POLY:
            g.setStroke(style.getStroke());
            Drawer.drawDecoratedPolyLine(g, getPoints(),
                    style.getLineStartDecoration(),
                    style.getLineEndDecoration());
            break;

        case POLY_SPLINED:
            g.setStroke(style.getStroke());
            Drawer.drawDecoratedSpline(g, getPoints(),
                    style.getLineStartDecoration(),
                    style.getLineEndDecoration());
            g.setStroke(Globals.LINE_STROKE_PHANTOM);
            g.setColor(Globals.LINE_COLOR_PANTOM);
            Drawer.drawDecoratedPolyLine(g, getPoints());
            break;

        case STRAIGHT:
            g.setStroke(style.getStroke());
            Drawer.drawDecoratedStraightLine(g, getPoints(),
                    style.getLineStartDecoration(),
                    style.getLineEndDecoration());
            break;
    }

}

@Override
protected void drawChildren(Graphics2D g) {

    double scale = g.getTransform().getScaleX();
    if (scale < 0.5)
        return;

    lv.lumii.layoutengine.Line lel = getLayoutElement();

    for (LineLabel lbl : lel.getLabels()) {
        DLineLabel mel = (DLineLabel) getThis(lbl);
        mel.draw(g);
    }

}

/**
 * Adds a text at the line start.
 *
 * @param text        Text to be added.
 * @param orientation Text placement relative to the line.
 * @return Label containing the given text.
 */
public DLineLabel addTextAtStart(String text, LineLabel.Orientation orientation) {
    DLineLabel lbl = createLabel(0, 0, 0, orientation);
    lbl.setRotation(true);
    lbl.setAutoFlip(true);
    lbl.addText(text);
    return lbl;
}

/**
 * Adds a text at the line start, placed on the left relative to the line.
 *
 * @param text Text to be added.
 * @return Label containing the given text.
 */
public DLineLabel addTextAtStart(String text) {
    return addTextAtStart(text, LineLabel.Orientation.CLOCKWISE);
}

/**
 * Adds a text at the line end.
 *
 * @param text        Text to be added.
 * @param orientation Text placement relative to the line.
 * @return Label containing the given text.
 */
public DLineLabel addTextAtEnd(String text, LineLabel.Orientation orientation) {
    DLineLabel lbl = createLabel(0, 0, 1, orientation);
    lbl.setRotation(true);
    lbl.setAutoFlip(true);
    lbl.addText(text);
    return lbl;
}

/**
 * Adds a text at the line end, placed on the right relative to the line.
 *
 * @param text Text to be added.
 * @return Label containing the given text.
 */
public DLineLabel addTextAtEnd(String text) {
    return addTextAtEnd(text, LineLabel.Orientation.CLOCKWISE);
}

/**
 * Adds a text at the line middle, placed centered relative to the line.
 *
 * @param text Text to be added.
 * @return Label containing the given text.
 */
public DLineLabel addTextAtMiddle(String text) {
    return addTextAtMiddle(text, LineLabel.Orientation.CENTER);
}

/**
 * Adds a text at the line middle.
 *
 * @param text        Text to be added.
 * @param orientation Text placement relative to the line.
 * @return Label containing the given text.
 */
public DLineLabel addTextAtMiddle(String text, LineLabel.Orientation orientation) {
    DLineLabel lbl = createLabel(0, 0, 0.5, orientation);
    lbl.setRotation(true);
    lbl.setAutoFlip(true);
    lbl.addText(text);
    return lbl;
}

@Override
public boolean isSelfHit(Point2D.Double p) {
    switch (_lineType) {
        case ORTHO:
        case ORTHO_SPLINED:
            return HitTest.checkOrthoLineHit(
                    getPoints(),
                    p.x, p.y);
        case POLY:
        case POLY_SPLINED:
            return HitTest.checkPolyLineHit(
                    getPoints(),
                    p.x, p.y);
        case STRAIGHT:
            return HitTest.checkFreeSegmentHit(
                    this.getStartPoint(),
                    this.getEndPoint(),
                    p.x, p.y);
    }
    return false;
}

@Override
public boolean isChildHit(Point2D.Double p) {

    for (DLineLabel lbl : labels()) {
        if (lbl.isHit(p))
            return true;
    }

    return false;
}

@Override
public DElement getChildHit(Point2D.Double p) {

    for (DLineLabel lbl : labels()) {
        DElement e = lbl.getHit(p);
        if (e != null)
            return e;

    }

    return null;
}

@Override
protected void drawTestInfo(Graphics2D g) {
    lv.lumii.layoutengine.Line lel = getLayoutElement();

    if (log.isLoggingOn()) {
        g.setFont(Globals.infoFont);
        g.drawString(String.format("#%d: %s",
                getThisId(lel),
                log.dumpLine(lel)),
                (float) lel.getStartPoint().getX(), (float) lel.getStartPoint().getY());
    }

}

@Override
public void recalculate() {
    for (DLineLabel lbl : labels())
        lbl.recalculate();
}

@Override
public void removeChildren() {
    for (DLineLabel l : labels()) {
        l.remove();
    }
}

/**
 *
 */
public void retrace() {
    getLayoutElement().retrace();
}

public void retrace(DLineType dLineType) {
    getLayoutElement().retrace(dLineType.getInternalType());
}

public void retrace(DBox newStart, DBox newEnd) {
    getLayoutElement().retrace(newStart.getLayoutElement(), newEnd.getLayoutElement());
}

public void retrace(DLineType dLineType, DBox newStart, DBox newEnd) {
    getLayoutElement().retrace(dLineType.getInternalType(), newStart.getLayoutElement(), newEnd.getLayoutElement());
}

public void cleanup() {
    if (getInternalLineType() != Line.LineType.ORTHOGONAL)
        return;
    getLayoutElement().cleanup();
}

public Iterable<DLineLabel> labels() {
    return new Iterable<DLineLabel>() {
        @Override
        public Iterator<DLineLabel> iterator() {
            return new Iterator<DLineLabel>() {
                private ArrayList<LineLabel> _items = getLayoutElement().getLabels();
                private int pos = 0;

                @Override
                public boolean hasNext() {
                    return pos < _items.size();
                }

                @Override
                public DLineLabel next() {
                    return getThisLineLabel(_items.get(pos++));
                }

                @Override
                public void remove() {

                }
            };
        }
    };
}

/**
 * Returns the mask of the allowed start sides for this box. The mask bits are as follows:
 * <pre>
 * 0b0001: top
 * 0b0010: right
 * 0b0100: bottom
 * 0b1000: left
 * </pre>
 *
 * @return the mask of the allowed start sides for this box.
 */
public int getStartSides() {
    return getLayoutElement().getStartSides();
}

/**
 * Sets the sides of the start box this line is allowed to start from. The sides are given as an
 * array of {@code BoxSides}, at least one side must be allowed.
 *
 * @param boxSides
 */
public void setStartSides(Box.BoxSide[] boxSides) {
    getLayoutElement().setStartSides(boxSides);
}

/**
 * Returns the mask of the allowed end sides for this box. The mask bits are as follows:
 * <pre>
 * 0b0001: top
 * 0b0010: right
 * 0b0100: bottom
 * 0b1000: left
 * </pre>
 *
 * @return the mask of the allowed end sides for this box.
 */
public int getEndSides() {
    return getLayoutElement().getEndSides();
}

/**
 * Sets the sides of the end box this line is allowed to end at. The sides are given as an
 * array of {@code BoxSides}, at least one side must be allowed.
 *
 * @param boxSides
 */
public void setEndSides(Box.BoxSide[] boxSides) {
    getLayoutElement().setEndSides(boxSides);
}

/**
 * Resets the start sides of this line to allow it to start from any side of the start box.
 */
public void resetStartSides() {
    getLayoutElement().resetStartSides();
}

/**
 * Resets the end sides of this line to allow it to end at any side of the end box.
 */
public void resetEndSides() {
    getLayoutElement().resetEndSides();
}

/**
 * Sets the sides of the start box this line is allowed to start from. The sides are given as an
 * mask, at least one side must be allowed. The mask bits are as follows:
 * <pre>
 * 0b0001: top
 * 0b0010: right
 * 0b0100: bottom
 * 0b1000: left
 * </pre>
 *
 * @param sides the mask of allowed start box sides for this line
 */
public void setStartSides(int sides) {
    getLayoutElement().setStartSides(sides);
}

/**
 * Sets the sides of the end box this line is allowed to end at. The sides are given as a mask,
 * at least one side must be allowed. The mask bits are as follows:
 * <pre>
 * 0b0001: top
 * 0b0010: right
 * 0b0100: bottom
 * 0b1000: left
 * </pre>
 *
 * @param sides the mask of allowed end box sides for this line
 */
public void setEndSides(int sides) {
    getLayoutElement().setEndSides(sides);
}

/**
 * Returns the context menu for this line.
 *
 * @return the context menu
 */
@Override
public JPopupMenu getElementMenu() {
    JPopupMenu menu = super.getElementMenu();

    menu.addSeparator();

    menu.add(menuCommand(Commands.SIMPLIFY_LINE, Commands.SIMPLIFY_LINE_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            cleanup();
            updateAll();
        }
    }));

    menu.add(menuCommand(Commands.RETRACE_LINE, Commands.RETRACE_LINE_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            retrace();
            updateAll();
        }
    }));

    menu.addSeparator();

    menu.add(getLineTypeSubmenu(Commands.LINE_TYPE_MENU, Commands.LINE_TYPE_PREFIX, getLineType(), new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            String actionCommand = e.getActionCommand();
            setLineType(DLineType.valueOf(actionCommand));
            updateAll();
        }
    }));

    menu.add(getLineSidesSubmenu(Commands.LINE_START_SIDES_MENU, Commands.LINE_START_SIDES_PREFIX, getStartSides(), new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            String actionCommand = e.getActionCommand();
            setStartSides(Integer.parseInt(actionCommand));
            updateAll();
        }
    }));

    menu.add(getLineSidesSubmenu(Commands.LINE_END_SIDES_MENU, Commands.LINE_END_SIDES_PREFIX, getEndSides(), new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            String actionCommand = e.getActionCommand();
            setEndSides(Integer.parseInt(actionCommand));
            updateAll();
        }
    }));

    return menu;
}

private JMenu getLineSidesSubmenu(String title, String prefix, int lineSides, ActionListener listener) {

    JMenuItem item;
    JMenu sidesSubmenu = new JMenu(title);

    for (int i = 15; i > 0; i--) {
        item = new JCheckBoxMenuItem(getLineSidesText(i));
        if (lineSides == i)
            item.setSelected(true);
        item.addActionListener(listener);
//        item.setActionCommand(prefix + ':' + Integer.toString(i));
        item.setActionCommand(Integer.toString(i));
        sidesSubmenu.add(item);
    }
    return sidesSubmenu;
}

private String getLineSidesText2(int lineSides) {
    return "sides: " + Integer.toString(lineSides, 2);
}

private String getLineSidesText(int lineSides) {
    String text = "";

    if ((lineSides & 0b0001) != 0)
        text += "N";
    else
        text += ".";

    if ((lineSides & 0b0010) != 0)
        text += "E";
    else
        text += ".";

    if ((lineSides & 0b0100) != 0)
        text += "S";
    else
        text += ".";

    if ((lineSides & 0b1000) != 0)
        text += "W";
    else
        text += ".";

    return text;
}

@Override
protected void saveChildrenToXML(Element e, Document doc, LinkedHashMap<Object, Integer> idMap) {
    super.saveChildrenToXML(e, doc, idMap);
    org.w3c.dom.Element eLineLabels = doc.createElement(XMLHelper.LINE_LABELS_ELEMENT);
    for (DLineLabel lbl : labels()) {
        eLineLabels.appendChild(lbl.saveToXML(doc, idMap));
    }
    e.appendChild(eLineLabels);
}


public enum DLineType {
    /**
     * Straight line
     */
    STRAIGHT,

    /**
     * List of ortho-linear segments
     */
    ORTHO,

    /**
     * Same as ORTHO, but the line will be (visually) approximated by a (cubic) spline.
     */
    ORTHO_SPLINED,

    /**
     * List of arbitrary oriented segments.
     */
    POLY,

    /**
     * Same as POLY, but the line will be (visually) approximated by a (cubic) spline.
     */
    POLY_SPLINED;

    /**
     * Finds the corresponding line type for the given internal line type.
     *
     * @param internalType
     * @return default line type for this internal type
     */
    static DLineType getDLineType(Line.LineType internalType) {
        switch (internalType) {
            case ORTHOGONAL:
                return ORTHO;

            case POLYLINE:
                return POLY;

            case STRAIGHT:
            default:
                return STRAIGHT;
        }
    }

    /**
     * Internal type of the line as used by the layout engine.
     *
     * @return internal type for this line type
     */
    Line.LineType getInternalType() {
        switch (this) {
            case ORTHO:
            case ORTHO_SPLINED:
                return Line.LineType.ORTHOGONAL;

            case POLY:
            case POLY_SPLINED:
                return Line.LineType.POLYLINE;

            case STRAIGHT:
                return Line.LineType.STRAIGHT;
        }
        return Line.LineType.ORTHOGONAL;
    }
}

private static class Commands {
    // do not use ':' in commands and prefixes, as it is used as separator here

    static final String SIMPLIFY_LINE = "Simplify";
    static final String SIMPLIFY_LINE_COMMAND = "line-simplify";
    static final String RETRACE_LINE = "Retrace";
    static final String RETRACE_LINE_COMMAND = "line-retrace";
    static final String LINE_TYPE_MENU = "Line Type";
    static final String LINE_TYPE_PREFIX = "line-settype";
    static final String LINE_START_SIDES_MENU = "Start sides";
    static final String LINE_START_SIDES_PREFIX = "line-setstartsides";
    static final String LINE_END_SIDES_MENU = "End sides";
    static final String LINE_END_SIDES_PREFIX = "line-setendsides";

}

}
