/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;

import lv.lumii.layoutengine.OutsideLabel;

import javax.swing.*;

/**
 * Line label is a label which is attached to a {@code DLine}.
 * <p/>
 * Line labels have a position along the line trace (from the start to the end),
 * as well as a position relative to the line path (to the right, directly on the path, etc.)
 * <p/>
 * Line labels can be manually <em>flipped</em> (rotated by 90 degrees) and <em>auto-flipped</em> (automatically flipped,
 * if changes the direction of the line segment where the label is located).
 * <p/>
 * For straight lines and poly lines, the labels can also be rotated, following the angle of the line segment.
 * <p/>
 * When diagram is being rearranged, all line labels are placed without overlapping, if possible.
 */
public class DLineLabel extends DOutsideLabel {


public DLineLabel(OutsideLabel.LineLabel label) {
    super(label);
}

DLineLabel(org.w3c.dom.Element element, LinkedHashMap<Integer, Object> objMap) {
    super(element, objMap);
}

@Override
public lv.lumii.layoutengine.OutsideLabel.LineLabel getLayoutElement() {
    return (lv.lumii.layoutengine.OutsideLabel.LineLabel) layoutElement;
}

@Override
protected void drawSelf(Graphics2D g) {

    lv.lumii.layoutengine.OutsideLabel.LineLabel lbl = getLayoutElement();
    Rectangle2D.Double r = getBounds();

    AffineTransform oldTransform = null;

    oldTransform = g.getTransform();

    if (lbl.isRotated()) {
        g.translate(r.getCenterX(), r.getCenterY());
        double angle = lbl.getRotationAngle();

        if (lbl.isFlipped()) { // flipped and rotated
            angle -= Math.PI / 2;
            g.rotate(angle);
            g.translate(-r.getHeight() / 2, -r.getWidth() / 2);
        } else { // just rotated
            g.rotate(angle);
            g.translate(-r.getWidth() / 2, -r.getHeight() / 2);
        }

    } else if (lbl.isFlipped()) { // just flipped
        double angle = -Math.PI / 2;
        g.translate(r.getX(), r.getY() + r.getHeight());
        g.rotate(angle);
    } else { // neither flipped not rotated
        g.translate(r.getX(), r.getY());
    }


    drawContent(g);

    g.setTransform(oldTransform);


}

public OutsideLabel.LineLabel.Orientation getOrientation() {
    return getLayoutElement().getOrientation();
}

public void setOrientation(OutsideLabel.LineLabel.Orientation orientation) {
    getLayoutElement().setOrientation(orientation);
}

public boolean isRotated() {
    return getLayoutElement().isRotated();
}

public void setRotation(boolean b) {
    getLayoutElement().setRotation(b);
}

public void move(Point2D.Double point, DLine owner) {
    getLayoutElement().move(point, owner.getLayoutElement());
}

public void move(double position, DLine owner) {
    getLayoutElement().move(position, owner.getLayoutElement());
}

@Override
public DLine getOwner() {
    return getThisLine(getLayoutElement().getOwner());
}

@Override
public JPopupMenu getElementMenu() {
    JPopupMenu menu = super.getElementMenu();

    menu.addSeparator();

    // ...

    menu.add(getLineLabelOrientationSubmenu(
            Commands.LINE_LABEL_ORIENTATION_MENU,
            Commands.LINE_LABEL_ORIENTATION_PREFIX,
            getOrientation(),
            new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    String actionCommand = e.getActionCommand();
                    setOrientation(OutsideLabel.LineLabel.Orientation.valueOf(actionCommand));
                    updateAll();
                }
            }));

    menu.add(menuCommand(Commands.LINE_LABEL_ROTATE, Commands.LINE_LABEL_ROTATE_COMMAND, new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            setRotation(!isRotated());
            updateAll();
        }
    },
            isRotated()
    ));

    return menu;
}

private JMenu getLineLabelOrientationSubmenu(String title, String prefix, OutsideLabel.LineLabel.Orientation labelOrientation, ActionListener listener) {

    JMenuItem item;
    JMenu orientationSubmenu = new JMenu(title);

    for (OutsideLabel.LineLabel.Orientation orient : OutsideLabel.LineLabel.Orientation.values()) {

        item = new JCheckBoxMenuItem(orient.toString());
        if (orient == labelOrientation)
            item.setSelected(true);
        item.addActionListener(listener);
//            item.setActionCommand(prefix + ':' + orient.toString());
        item.setActionCommand(orient.toString());
        orientationSubmenu.add(item);
    }
    return orientationSubmenu;
}

private static class Commands {
    // do not use ':' in commands and prefixes, as it is used as separator here

    static final String LINE_LABEL_ORIENTATION_MENU = "Position";
    static final String LINE_LABEL_ORIENTATION_PREFIX = "linelabel-setorientation";
    static final String LINE_LABEL_ROTATE = "Rotate";
    static final String LINE_LABEL_ROTATE_COMMAND = "outsidelabel-setrotate";

}
}
