/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.model;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;

//import org.w3c.dom.*;

/**
 * Stores common methods for saving and loading diagrams to XML.
 */
public abstract class XMLHelper {

/**
 * The XML namespace used for diagram storage.
 */
public static final String NAMESPACE = "mg";
public static final String LAYOUT_NAMESPACE = "jk";
public static final String DRAWING_ELEMENT = NAMESPACE + ":drawing";
public static final String DIAGRAM_ELEMENT = NAMESPACE + ":diagram";
public static final String BOXES_ELEMENT = NAMESPACE + ":boxes";
public static final String BOX_ELEMENT = NAMESPACE + ":box";
public static final String LINES_ELEMENT = NAMESPACE + ":lines";
public static final String LINE_ELEMENT = NAMESPACE + ":line";
public static final String GROUPS_ELEMENT = NAMESPACE + ":groups";
public static final String GROUP_ELEMENT = NAMESPACE + ":group";
public static final String INSIDE_LABELS_ELEMENT = NAMESPACE + ":insidelabels";
public static final String INSIDE_LABEL_ELEMENT = NAMESPACE + ":insidelabel";
public static final String BOX_LABELS_ELEMENT = NAMESPACE + ":boxlabels";
public static final String BOX_LABEL_ELEMENT = NAMESPACE + ":boxlabel";
public static final String LINE_LABELS_ELEMENT = NAMESPACE + ":linelabels";
public static final String LINE_LABEL_ELEMENT = NAMESPACE + ":linelabel";
public static final String STYLE_ELEMENT = NAMESPACE + ":style";
public static final String DATA_ELEMENT = NAMESPACE + ":data";
public static final String DATA_ITEM = NAMESPACE + ":dataitem";
public static final String PACK_ELEMENT = NAMESPACE + ":pack";
public static final String TEXTBLOCK_ELEMENT = NAMESPACE + ":textblock";
public static final String SEPARATOR_ELEMENT = NAMESPACE + ":separator";
public static final String LAYOUT_ELEMENT = LAYOUT_NAMESPACE + ":diagram";
/**
 * The URI of the diagram namespace. TODO: Use a real URI.
 */
public static String NAMESPACE_URI = "http://www.lumii.lv/mg";

/**
 * Save the given rectangle to XML.
 *
 * @param doc       the XML document in which to create the new XML element
 * @param rectangle the rectangle from which to create an XML element
 * @param name      the name of the created XML element
 * @return an XML element corresponding to the given rectangle
 */
static org.w3c.dom.Element save(org.w3c.dom.Document doc, Rectangle2D.Double rectangle, String name) {
    org.w3c.dom.Element e = doc.createElement(name);

    e.setAttribute("x", Double.toString(rectangle.x));
    e.setAttribute("y", Double.toString(rectangle.y));
    e.setAttribute("width", Double.toString(rectangle.width));
    e.setAttribute("height", Double.toString(rectangle.height));

    return e;
}

/**
 * Save the given point to XML.
 *
 * @param doc   the XML document in which to create the new XML element
 * @param point the point from which to create an XML element
 * @param name  the name of the created XML element
 * @return an XML element corresponding to the given point
 */
static org.w3c.dom.Node save(org.w3c.dom.Document doc, Point2D.Double point, String name) {
    org.w3c.dom.Element e = doc.createElement(name);

    e.setAttribute("x", Double.toString(point.x));
    e.setAttribute("y", Double.toString(point.y));

    return e;
}

/**
 * Loads a rectangle from the given XML element
 *
 * @param element an XML element corresponding to a rectangle
 * @return the rectangle created from the given XML element
 */
static Rectangle2D.Double loadRectangle(org.w3c.dom.Element element) {
    double x, y, w, h;
    x = Double.valueOf(element.getAttribute("x"));
    y = Double.valueOf(element.getAttribute("y"));
    w = Double.valueOf(element.getAttribute("width"));
    h = Double.valueOf(element.getAttribute("height"));
    return new Rectangle2D.Double(x, y, w, h);
}

/**
 * Loads a point from the given XML element
 *
 * @param element an XML element corresponding to a point
 * @return the point created from the given XML element
 */
static Point2D.Double loadPoint(org.w3c.dom.Element element) {
    double x, y;
    x = Double.valueOf(element.getAttribute("x"));
    y = Double.valueOf(element.getAttribute("y"));
    return new Point2D.Double(x, y);
}
//public static final String DATA_KEY = NAMESPACE + ":dataitem";
//public static final String DATA_VALUE = NAMESPACE + ":dataitem";

/**
 * Loads a Double attribute from the given XML element
 *
 * @param e    an XML element with a double attribute
 * @param name the name of the attribute to load
 * @return a Double value corresponding to the given attribute, or {@code Double.NaN} if there
 * isn't such an attribute.
 */
static double loadDouble(org.w3c.dom.Element e, String name) {
    return e.hasAttribute(name) ? Double.parseDouble(e.getAttribute(name)) : Double.NaN;
}

/**
 * Loads an integer attribute from the given XML element
 *
 * @param e    an XML element with an integer attribute
 * @param name the name of the attribute to load
 * @return an integer value corresponding to the given attribute, or 0 if there isn't such an
 * attribute.
 */
static int loadInt(org.w3c.dom.Element e, String name) {
    return e.hasAttribute(name) ? Integer.parseInt(e.getAttribute(name)) : 0;
}

/**
 * Loads a long integer from the given XML element
 *
 * @param e    an XML element with a long attribute
 * @param name the name of the attribute to load
 * @return a long value corresponding to the given attribute, or 0 if there isn't such an
 * attribute.
 */
static long loadLong(org.w3c.dom.Element e, String name) {
    return e.hasAttribute(name) ? Long.parseLong(e.getAttribute(name)) : 0L;
}

static void buildChildren(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objMap) {
}

}
