/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.style;

import org.w3c.dom.Element;
import org.w3c.dom.Document;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;

public interface Style {

/**
 * Returns color to be used for lines and for shape borders.
 *
 * @return stroke Color
 */
Color getStrokeColor();

/**
 * Sets the color to be used for lines and for shape borders.
 *
 * @param color
 * @return this style
 */
Style setStrokeColor(Color color);

/**
 * Returns the color to be used for inside area fills.
 * If gradient is used, this is one of two colors determining the gradient.
 *
 * @return Color
 */
Color getFillColor();

/**
 * Sets the color to be used for inside area fills.
 * If gradient is used, this is one of two colors determining the gradient.
 *
 * @param color
 * @return this style
 */
Style setFillColor(Color color);

/**
 * Returns color to be used for text outputs.
 *
 * @return Color
 */
Color getTextColor();

/**
 * Sets the color to be used for text outputs.
 *
 * @param color
 * @return this style
 */
Style setTextColor(Color color);

/**
 * Returns the secondary color for gradient fills.
 * Gradient ranges from the primary fill color to this color.
 *
 * @return gradient Color
 */
Color getGradientColor();

/**
 * Sets the secondary color for gradient fills.
 * Gradient ranges from the primary fill color to this color.
 *
 * @param color
 * @return this style
 */
Style setGradientColor(Color color);

/**
 * Returns the currently used gradient type.
 *
 * @return gradient type
 */
GradientType getGradientType();

/**
 * Sets the currently used gradient type.
 *
 * @param gradientType
 * @return this style
 */
Style setGradientType(GradientType gradientType);

/**
 * Returns the paint used for inside area fills.
 * If gradient is used then both fill color and gradient color is used.
 * Without gradient, fill color is returned.
 *
 * @return Paint
 */
Paint getPaint();

/**
 * Returns the paint used for inside area fills.
 * If gradient is used then both fill color and gradient color is used.
 * Without gradient, fill color is returned.
 *
 * @param rect rectangle for gradient calculations
 * @return Paint
 */
Paint getPaint(Rectangle2D.Double rect);

/**
 * Returns the font to be used in this style
 *
 * @return Font
 */
Font getFont();

/**
 * Returns the typeface to be used in the style font.
 *
 * @return typeface name
 */
String getFontTypeface();

/**
 * Set the typeface to be used in the style font.
 *
 * @param typeface
 * @return this style
 */
Style setFontTypeface(String typeface);

/**
 * Returns the font size to be used in this style
 *
 * @return font size
 */
int getFontSize();

/**
 * Sets the font size to be used in this style
 *
 * @param size font size
 * @return this style
 */
Style setFontSize(int size);

/**
 * Returns weight of the style font
 *
 * @return true if bold
 */
boolean isFontBold();

/**
 * Sets the weight of the style font
 *
 * @param isBold true if bold
 * @return this style
 */
Style setFontBold(boolean isBold);

/**
 * Returns the italics of the style font
 *
 * @return true if italics
 */
boolean isFontItalics();

/**
 * Sets the italics of the style font
 *
 * @param isItalics
 * @return this style
 */
Style setFontItalics(boolean isItalics);

/**
 * Returns the stroke width of this style.
 *
 * @return stroke width
 */
float getStrokeWidth();

/**
 * Sets the stroke width for lines drawn with this style
 *
 * @param width
 * @return this style
 */
Style setStrokeWidth(float width);

/**
 * Returns the end caps to be used for lines drawn with this style
 *
 * @return end caps
 */
int getStrokeEndCap();

/**
 * Sets the end caps to be used for lines drawn with this style
 *
 * @param endCap
 * @return this style
 */
Style setStrokeEndCap(int endCap);

/**
 * Returns the join type to be used for lines drawn with this style
 *
 * @return join type
 */
int getStrokeJoin();

/**
 * Sets the join type to be used for lines drawn with this style
 *
 * @param join join type
 * @return this style
 */
Style setStrokeJoin(int join);

/**
 * Returns the miter limit to be used for lines drawn with this style
 *
 * @return miter limit
 */
float getStrokeMiterLimit();

/**
 * Sets the miter limit to be used for lines drawn with this style
 *
 * @param miterLimit miter limit
 * @return this style
 */
Style setStrokeMiterLimit(float miterLimit);

/**
 * Returns the dash array to be used for lines drawn with this style
 *
 * @return the dash array
 */
float[] getStrokeDash();

/**
 * Sets the dash array to be used for lines drawn with this style
 *
 * @param dash dash array
 * @return this style
 */
Style setStrokeDash(float[] dash);

/**
 * Returns the dash phase to be used for lines drawn with this style
 *
 * @return dash phase
 */
float getStrokeDashPhase();

/**
 * Sets the dash phase to be used for lines drawn with this style
 *
 * @param dashPhase dash phase
 * @return this style
 */
Style setStrokeDashPhase(float dashPhase);

/**
 * Returns the stroke to be used for lines and borders drawn with this style.
 *
 * @return Stroke
 */
Stroke getStroke();

/**
 * Returns the decoration type to be used at the line start point
 *
 * @return decoration type
 */
LineDecorationType getLineStartDecoration();

/**
 * Sets the decoration type to be used at the line start point
 *
 * @param decorationType decoration type
 * @return this style
 */
Style setLineStartDecoration(LineDecorationType decorationType);

/**
 * Returns the decoration type to be used at the line end point
 *
 * @return decoration type
 */
LineDecorationType getLineEndDecoration();

/**
 * Sets the decoration type to be used at the line end point
 *
 * @param decorationType
 * @return this style
 */
Style setLineEndDecoration(LineDecorationType decorationType);

/**
 * Returns the box shape type for this element
 *
 * @return box shape type
 */
BoxShapeType getBoxShapeType();

/**
 * Sets the box shape type for this element
 *
 * @param boxShapeType box shape type
 * @return this style
 */
Style setBoxShapeType(BoxShapeType boxShapeType);

//double getSpacing();
//Style setSpacing(double spacing);

/**
 * Returns the padding to be used for this element
 *
 * @return padding
 */
double getPadding();

/**
 * Sets the padding to be used for this element
 *
 * @param padding padding
 * @return this style
 */
Style setPadding(double padding);

/**
 * Whether this element has a shadow
 *
 * @return whether this element has a shadow
 */
boolean hasShadow();

/**
 * Sets the shadow for this element.
 *
 * @param shadow true if shadow is to be used
 * @return this style
 */
Style setShadow(boolean shadow);

/**
 * Saves this element to a xml element and returns it.
 * If needed, the child elements are also included in this element.
 *
 * @param doc org.w3c.dom.Document
 * @return
 */
Element saveToXML(Document doc);

//Style clone();

JMenu getMenu(StyleMenuType menuType);

void delayUpdate(boolean inUpdate);

enum StyleMenuType {ELEMENT, LINE, BOX, TEXT}

//    Style(org.w3c.dom.Element e, LinkedHashMap<Integer, Object> objectMap, Element owner);
//    org.w3c.dom.Element saveToXML(org.w3c.dom.Document doc, LinkedHashMap<Object, Integer> idMap);
}
