/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.style;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Rectangle2D;

import lv.lumii.deek.model.DElement;
import lv.lumii.deek.model.XMLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.swing.*;

public class StyleImpl implements Style {

private static final Color defaultStrokeColor = Color.BLACK;
private static final Color defaultFillColor = Color.GRAY;
private static final Color defaultTextColor = Color.BLACK;
private static final Color defaultGradientColor = Color.WHITE;
private static final GradientType defaultGradientType = GradientType.TOP_TO_BOTTOM;
private static final int defaultFontSize = 16;
private static final String defaultFontTypeface = Font.SANS_SERIF;
private static final boolean defaultFontBold = false;
private static final boolean defaultFontItalics = false;
private static final float defaultStrokeWidth = 1.0f;
private static final int defaultStrokeEndCap = BasicStroke.CAP_ROUND;
private static final int defaultStrokeJoin = BasicStroke.JOIN_ROUND;
private static final float defaultStrokeMiterLimit = 1.0f;
private static final float defaultStrokeDashPhase = 0.0f;
private static final LineDecorationType defaultLineStartDecoration = LineDecorationType.NONE;
private static final LineDecorationType defaultLineEndDecoration = LineDecorationType.FILLED_ARROW;
private static final BoxShapeType defaultBoxShapeType = BoxShapeType.RECTANGLE;
private static final double defaultPadding = 3;
private static final boolean defaultShadow = true;
private final static String STROKE_COLOR_ATTR = "strokecolor";
private final static String FILL_COLOR_ATTR = "fillcolor";
private final static String GRADIENT_COLOR_ATTR = "gradientcolor";
private final static String GRADIENT_TYPE_ATTR = "gradienttype";
private final static String TEXT_COLOR_ATTR = "textcolor";
private final static String FONT_SIZE_ATTR = "fontsize";
private final static String FONT_FACE_ATTR = "fontface";
private final static String FONT_BOLD_ATTR = "fontbold";
private final static String FONT_ITALICS_ATTR = "fontitalics";
private final static String STROKE_WIDTH_ATTR = "strokewidth";
private final static String STROKE_ENDCAP_ATTR = "strokeendcap";
private final static String STROKE_JOIN_ATTR = "strokejoin";
private final static String STROKE_MITER_LIMIT_ATTR = "strokemiterlimit";
private final static String STROKE_DASH_ATTR = "strokedash";
private final static String STROKE_DASH_PHASE_ATTR = "strokedashphase";
private final static String LINE_START_DECORATION_ATTR = "linestartdecoration";
private final static String LINE_END_DECORATION_ATTR = "lineenddecoration";
private final static String BOX_SHAPE_ATTR = "boxshape";
private final static String PADDING_ATTR = "padding";
private final static String SHADOW_ATTR = "shadow";
private DElement owner;
private Color strokeColor = defaultStrokeColor;
private Color fillColor = defaultFillColor;
private Color textColor = defaultTextColor;
private Color gradientColor = defaultGradientColor;
private GradientType gradientType = defaultGradientType;
private Font font;
private int fontSize = defaultFontSize;
private String fontTypeface = defaultFontTypeface;
private boolean fontBold = defaultFontBold;
private boolean fontItalics = defaultFontItalics;
private float strokeWidth = defaultStrokeWidth;
private int strokeEndCap = defaultStrokeEndCap;
private int strokeJoin = defaultStrokeJoin;
private float strokeMiterLimit = defaultStrokeMiterLimit;
private float[] strokeDash;
private float strokeDashPhase = defaultStrokeDashPhase;
private Stroke stroke = null;
private LineDecorationType lineStartDecoration = defaultLineStartDecoration;
private LineDecorationType lineEndDecoration = defaultLineEndDecoration;
private BoxShapeType boxShapeType = defaultBoxShapeType;
private double padding = defaultPadding;
private boolean shadow = defaultShadow;
private boolean delayUpdate = false;

public StyleImpl(DElement element) {
    owner = element;
}

public StyleImpl() {
    owner = null;
}

public StyleImpl(DElement element, Element e) {
    owner = element;
    loadAttributesFromXML(e);
}

@Override
public Color getStrokeColor() {
    return strokeColor;
}

@Override
public Style setStrokeColor(Color color) {
    if (this.strokeColor != color) {
        this.strokeColor = color;
        updateOwner();
    }
    return this;
}

protected JMenuItem getStrokeColorMenuItem() {
    return menuCommand(Commands.EDIT_STYLE_STROKECOLOR, Commands.EDIT_STYLE_STROKECOLOR_COMMAND,
            new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    Color newColor = JColorChooser.showDialog(
//                        _frame,
                            null,
                            Commands.EDIT_STYLE_STROKECOLOR,
                            Color.BLACK);
                    if (newColor != null) {
                        setStrokeColor(newColor);
                    }
                }
            });
}

@Override
public Color getFillColor() {
    return fillColor;
}

@Override
public Style setFillColor(Color color) {
    if (this.fillColor != color) {
        this.fillColor = color;
        updateOwner();
    }
    return this;
}

protected JMenuItem getFillColorMenuItem() {
    return menuCommand(Commands.EDIT_STYLE_FILLCOLOR, Commands.EDIT_STYLE_FILLCOLOR_COMMAND,
            new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    Color newColor = JColorChooser.showDialog(
//                        _frame,
                            null,
                            Commands.EDIT_STYLE_FILLCOLOR,
                            Color.BLACK);
                    if (newColor != null) {
                        setFillColor(newColor);
                    }
                }
            });
}

@Override
public Color getTextColor() {
    return textColor;
}

@Override
public Style setTextColor(Color color) {
    if (this.textColor != color) {
        this.textColor = color;
        updateOwner();
    }
    return this;
}

protected JMenuItem getTextColorMenuItem() {
    return menuCommand(Commands.EDIT_STYLE_TEXTCOLOR, Commands.EDIT_STYLE_TEXTCOLOR_COMMAND,
            new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    Color newColor = JColorChooser.showDialog(
//                        _frame,
                            null,
                            Commands.EDIT_STYLE_TEXTCOLOR,
                            Color.BLACK);
                    if (newColor != null) {
                        setTextColor(newColor);
                    }
                }
            });
}

@Override
public Style setGradientColor(Color color) {
    if (this.gradientColor != color) {
        this.gradientColor = color;
        updateOwner();
    }
    return this;
}

@Override
public Color getGradientColor() {
    return gradientColor;
}

protected JMenuItem getGradientColorMenuItem() {
    return menuCommand(Commands.EDIT_STYLE_GRADIENTCOLOR, Commands.EDIT_STYLE_GRADIENTCOLOR_COMMAND,
            new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    Color newColor = JColorChooser.showDialog(
//                        _frame,
                            null,
                            Commands.EDIT_STYLE_GRADIENTCOLOR,
                            Color.BLACK);
                    if (newColor != null) {
                        setGradientColor(newColor);
                    }
                }
            });
}

@Override
public GradientType getGradientType() {
    return gradientType;
}

@Override
public Style setGradientType(GradientType gradientType) {
    if (this.gradientType != gradientType) {
        this.gradientType = gradientType;
        updateOwner();
    }
    return this;
}

protected JMenu getGradientTypeSubmenu() {

    JMenuItem item;
    JMenu subMenu = new JMenu("Gradient type");

    for (final GradientType gt : GradientType.values()) {
        item = new JCheckBoxMenuItem(gt.toString());
        if (getGradientType() == gt)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setGradientType(gt);
            }
        });
        subMenu.add(item);
    }

    return subMenu;
}

@Override
public Paint getPaint() {
    return fillColor;
}

@Override
public Paint getPaint(Rectangle2D.Double r) {

    if (r == null)
        return fillColor;

    switch (gradientType) {

        case TOP_TO_BOTTOM:
            return new GradientPaint(
                    (float) r.x, (float) r.y,
                    fillColor,
                    (float) r.x, (float) (r.y + r.height),
                    gradientColor);

        case LEFT_TO_RIGHT:
            return new GradientPaint(
                    (float) r.x, (float) r.y,
                    fillColor,
                    (float) (r.x + r.width), (float) r.y,
                    gradientColor);

        case TOPLEFT_TO_BOTTOMRIGHT:
            return new GradientPaint(
                    (float) r.x, (float) r.y,
                    fillColor,
                    (float) (r.x + r.width), (float) (r.y + r.height),
                    gradientColor);

        case BOTTOMLEFT_TO_TOPRIGHT:
            return new GradientPaint(
                    (float) r.x, (float) (r.y + r.height),
                    fillColor,
                    (float) (r.x + r.width), (float) r.y,
                    gradientColor);

        case NONE:
        default:
            return fillColor;
    }

}

@Override
public Font getFont() {

    if (font != null)
        return font;

    int fontStyle = (fontBold ? Font.BOLD : Font.PLAIN)
            | (fontItalics ? Font.ITALIC : Font.PLAIN);

    font = new Font(fontTypeface, fontStyle, fontSize);

    return font;
}

@Override
public Style setFontSize(int size) {
    if (this.fontSize != size) {
        font = null;
        this.fontSize = size;
        updateOwner();
    }
    return this;
}

@Override
public int getFontSize() {
    return fontSize;
}

protected JMenu getFontSizeSubmenu() {
    JMenuItem item;
    JMenu menu = new JMenu("Font size");
    int[] fontSizes = {6, 8, 10, 12, 15, 18, 24, 32, 48};

    for (final int size : fontSizes) {
        item = new JCheckBoxMenuItem(Integer.toString(size));
        if (getFontSize() == size)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setFontSize(size);
            }
        });
        menu.add(item);
    }
    return menu;
}

@Override
public String getFontTypeface() {
    return fontTypeface;
}

@Override
public Style setFontTypeface(String fontTypeface) {
    if (!this.fontTypeface.equals(fontTypeface)) {
        font = null;
        this.fontTypeface = fontTypeface;
        updateOwner();
    }
    return this;
}

protected JMenu getFontTypefaceSubmenu() {
    JMenuItem item;
    JMenu menu = new JMenu("Typeface");

    String[] typefaces = {Font.SANS_SERIF, Font.SERIF, Font.MONOSPACED};

    for (final String face : typefaces) {
        item = new JCheckBoxMenuItem(face);
        if (getFontTypeface().equals(face))
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setFontTypeface(face);
            }
        });
        menu.add(item);
    }
    return menu;
}

@Override
public boolean isFontBold() {
    return fontBold;
}

@Override
public Style setFontBold(boolean useBold) {
    if (this.fontBold != useBold) {
        font = null;
        this.fontBold = useBold;
        updateOwner();
    }
    return this;
}

@Override
public boolean isFontItalics() {
    return fontItalics;
}

@Override
public Style setFontItalics(boolean useItalics) {
    if (this.fontItalics != useItalics) {
        font = null;
        this.fontItalics = useItalics;
        updateOwner();
    }
    return this;
}

protected JMenu getFontVariationsSubmenu() {
    JCheckBoxMenuItem item;
    JMenu menu = new JMenu("Font variations");

    item = new JCheckBoxMenuItem("Bold");
    if (isFontBold())
        item.setSelected(true);
    item.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            setFontBold(!isFontBold());
        }
    });
//    item.addItemListener(new ItemListener() {
//        @Override
//        public void itemStateChanged(ItemEvent e) {
//            setFontBold(e.getStateChange() == ItemEvent.SELECTED);
//        }
//    });
    menu.add(item);

    item = new JCheckBoxMenuItem("Italics");
    if (isFontItalics())
        item.setSelected(true);
    item.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            setFontItalics(!isFontItalics());
        }
    });
    menu.add(item);

    return menu;

}

@Override
public float getStrokeWidth() {
    return strokeWidth;
}

@Override
public Style setStrokeWidth(float width) {
    if (this.strokeWidth != width) {
        stroke = null;
        this.strokeWidth = width;
        updateOwner();
    }
    return this;
}

protected JMenu getStrokeWidthSubmenu() {
    JMenuItem item;
    JMenu menu = new JMenu("Stroke width");
    float[] widths = {0, 1, 2, 3, 4, 5};

    for (final float w : widths) {
        item = new JCheckBoxMenuItem(Float.toString(w));
        if (getStrokeWidth() == w)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setStrokeWidth(w);
            }
        });
        menu.add(item);
    }
    return menu;
}

@Override
public int getStrokeEndCap() {
    return strokeEndCap;
}

@Override
public Style setStrokeEndCap(int endCap) {
    if (this.strokeEndCap != endCap) {
        stroke = null;
        this.strokeEndCap = endCap;
        updateOwner();
    }
    return this;
}

@Override
public int getStrokeJoin() {
    return strokeJoin;
}

@Override
public Style setStrokeJoin(int join) {
    if (this.strokeJoin != join) {
        stroke = null;
        this.strokeJoin = join;
        updateOwner();
    }
    return this;
}

@Override
public float getStrokeMiterLimit() {
    return strokeMiterLimit;
}

@Override
public Style setStrokeMiterLimit(float miterLimit) {
    if (this.strokeMiterLimit != miterLimit) {
        stroke = null;
        this.strokeMiterLimit = miterLimit;
        updateOwner();
    }
    return this;
}

@Override
public float[] getStrokeDash() {
    return strokeDash;
}

@Override
public Style setStrokeDash(float[] dash) {
    strokeDash = dash;
    return this;
}

@Override
public float getStrokeDashPhase() {
    return strokeDashPhase;
}

@Override
public Style setStrokeDashPhase(float dashPhase) {
    if (this.strokeDashPhase != dashPhase) {
        stroke = null;
        this.strokeDashPhase = dashPhase;
        updateOwner();
    }
    return this;
}

@Override
public Stroke getStroke() {
    if (stroke != null)
        return stroke;

    if (strokeDash == null) {
        stroke = new BasicStroke(strokeWidth, strokeEndCap, strokeJoin, strokeMiterLimit);
    } else {
        stroke = new BasicStroke(strokeWidth, strokeEndCap, strokeJoin, strokeMiterLimit, strokeDash, strokeDashPhase);
    }

    return stroke;
}

@Override
public Style setLineStartDecoration(LineDecorationType decorationType) {
    if (this.lineStartDecoration != decorationType) {
        this.lineStartDecoration = decorationType;
        updateOwner();
    }
    return this;
}

@Override
public LineDecorationType getLineStartDecoration() {
    return lineStartDecoration;
}

@Override
public Style setLineEndDecoration(LineDecorationType decorationType) {
    if (this.lineEndDecoration != decorationType) {
        this.lineEndDecoration = decorationType;
        updateOwner();
    }
    return this;
}

@Override
public LineDecorationType getLineEndDecoration() {
    return lineEndDecoration;
}

protected JMenu getLineStartDecorationTypeSubmenu() {
    JMenuItem item;
    JMenu menu = new JMenu("Line Start");

    for (final LineDecorationType ldt : LineDecorationType.values()) {
        item = new JCheckBoxMenuItem(ldt.toString());
        if (getLineStartDecoration() == ldt)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setLineStartDecoration(ldt);
            }
        });
        menu.add(item);
    }

    return menu;
}

protected JMenu getLineEndDecorationTypeSubmenu() {
    JMenuItem item;
    JMenu menu = new JMenu("Line End");

    for (final LineDecorationType ldt : LineDecorationType.values()) {
        item = new JCheckBoxMenuItem(ldt.toString());
        if (getLineEndDecoration() == ldt)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setLineEndDecoration(ldt);
            }
        });
        menu.add(item);
    }

    return menu;
}

@Override
public Style setBoxShapeType(BoxShapeType boxShapeType) {
    if (this.boxShapeType != boxShapeType) {
        this.boxShapeType = boxShapeType;
        updateOwner();
    }
    return this;
}

@Override
public BoxShapeType getBoxShapeType() {
    return boxShapeType;
}

protected JMenu getBoxShapeTypeSubmenu() {
    JMenuItem item;
    JMenu menu = new JMenu("Shape type");

    for (final BoxShapeType bst : BoxShapeType.values()) {
        item = new JCheckBoxMenuItem(bst.toString());
        if (getBoxShapeType() == bst)
            item.setSelected(true);
        item.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setBoxShapeType(bst);
            }
        });
        menu.add(item);
    }
    return menu;
}

@Override
public double getPadding() {
    return padding;
}

@Override
public Style setPadding(double padding) {
    if (this.padding != padding) {
        this.padding = padding;
        updateOwner();
    }
    return this;
}

@Override
public boolean hasShadow() {
    return shadow;
}

@Override
public Style setShadow(boolean useShadow) {
    if (this.shadow != useShadow) {
        this.shadow = useShadow;
        updateOwner();
    }
    return this;
}

@Override
public Element saveToXML(Document doc) {
    Element e = doc.createElement(XMLHelper.STYLE_ELEMENT);
    saveAttributesToXML(e);
    return e;
}

private void saveAttributesToXML(Element e) {
    if (strokeColor != defaultStrokeColor) {
        e.setAttribute(STROKE_COLOR_ATTR, Integer.toString(strokeColor.getRGB()));
    }

    if (fillColor != defaultFillColor) {
        e.setAttribute(FILL_COLOR_ATTR, Integer.toString(fillColor.getRGB()));
    }

    if (gradientColor != defaultGradientColor) {
        e.setAttribute(GRADIENT_COLOR_ATTR, Integer.toString(gradientColor.getRGB()));
    }

    if (textColor != defaultTextColor) {
        e.setAttribute(TEXT_COLOR_ATTR, Integer.toString(textColor.getRGB()));
    }

    if (gradientType != defaultGradientType) {
        e.setAttribute(GRADIENT_TYPE_ATTR, gradientType.toString());
    }

    if (fontSize != defaultFontSize) {
        e.setAttribute(FONT_SIZE_ATTR, Integer.toString(fontSize));
    }

    if (!defaultFontTypeface.equals(fontTypeface)) {
        e.setAttribute(FONT_FACE_ATTR, fontTypeface);
    }

    if (fontBold != defaultFontBold) {
        e.setAttribute(FONT_BOLD_ATTR, Boolean.toString(fontBold));
    }

    if (fontItalics != defaultFontItalics) {
        e.setAttribute(FONT_ITALICS_ATTR, Boolean.toString(fontItalics));
    }

    if (strokeWidth != defaultStrokeWidth) {
        e.setAttribute(STROKE_WIDTH_ATTR, Float.toString(strokeWidth));
    }

    if (strokeEndCap != defaultStrokeEndCap) {
        e.setAttribute(STROKE_ENDCAP_ATTR, Integer.toString(strokeEndCap));
    }

    if (strokeJoin != defaultStrokeJoin) {
        e.setAttribute(STROKE_JOIN_ATTR, Integer.toString(strokeJoin));
    }

    if (strokeMiterLimit != defaultStrokeMiterLimit) {
        e.setAttribute(STROKE_MITER_LIMIT_ATTR, Float.toString(strokeMiterLimit));
    }

    if (strokeDashPhase != defaultStrokeDashPhase) {
        e.setAttribute(STROKE_DASH_PHASE_ATTR, Float.toString(strokeDashPhase));
    }

    if (lineStartDecoration != defaultLineStartDecoration) {
        e.setAttribute(LINE_START_DECORATION_ATTR, lineStartDecoration.toString());
    }

    if (lineEndDecoration != defaultLineEndDecoration) {
        e.setAttribute(LINE_END_DECORATION_ATTR, lineEndDecoration.toString());
    }

    if (boxShapeType != defaultBoxShapeType) {
        e.setAttribute(BOX_SHAPE_ATTR, boxShapeType.toString());
    }

    if (padding != defaultPadding) {
        e.setAttribute(PADDING_ATTR, Double.toString(padding));
    }

    if (shadow != defaultShadow) {
        e.setAttribute(SHADOW_ATTR, Boolean.toString(shadow));
    }


}

private void loadAttributesFromXML(Element e) {
    String tempStr;

    tempStr = e.getAttribute(STROKE_COLOR_ATTR);
    if (notEmpty(tempStr)) {
        strokeColor = new Color(Integer.parseInt(tempStr));
    }

    tempStr = e.getAttribute(FILL_COLOR_ATTR);
    if (notEmpty(tempStr)) {
        fillColor = new Color(Integer.parseInt(tempStr));
    }

    tempStr = e.getAttribute(GRADIENT_COLOR_ATTR);
    if (notEmpty(tempStr)) {
        gradientColor = new Color(Integer.parseInt(tempStr));
    }

    tempStr = e.getAttribute(TEXT_COLOR_ATTR);
    if (notEmpty(tempStr)) {
        textColor = new Color(Integer.parseInt(tempStr));
    }

    tempStr = e.getAttribute(GRADIENT_TYPE_ATTR);
    if (notEmpty(tempStr)) {
        gradientType = GradientType.valueOf(tempStr);
    }

    tempStr = e.getAttribute(FONT_SIZE_ATTR);
    if (notEmpty(tempStr)) {
        fontSize = Integer.parseInt(tempStr);
    }

    tempStr = e.getAttribute(FONT_FACE_ATTR);
    if (notEmpty(tempStr)) {
        fontTypeface = tempStr;
    }

    tempStr = e.getAttribute(FONT_BOLD_ATTR);
    if (notEmpty(tempStr)) {
        fontBold = Boolean.parseBoolean(tempStr);
    }

    tempStr = e.getAttribute(FONT_ITALICS_ATTR);
    if (notEmpty(tempStr)) {
        fontItalics = Boolean.parseBoolean(tempStr);
    }

    tempStr = e.getAttribute(STROKE_WIDTH_ATTR);
    if (notEmpty(tempStr)) {
        strokeWidth = Float.parseFloat(tempStr);
    }

    tempStr = e.getAttribute(STROKE_ENDCAP_ATTR);
    if (notEmpty(tempStr)) {
        strokeEndCap = Integer.parseInt(tempStr);
    }

    tempStr = e.getAttribute(STROKE_JOIN_ATTR);
    if (notEmpty(tempStr)) {
        strokeJoin = Integer.parseInt(tempStr);
    }

    tempStr = e.getAttribute(STROKE_MITER_LIMIT_ATTR);
    if (notEmpty(tempStr)) {
        strokeMiterLimit = Float.parseFloat(tempStr);
    }


    tempStr = e.getAttribute(STROKE_DASH_PHASE_ATTR);
    if (notEmpty(tempStr)) {
        strokeDashPhase = Float.parseFloat(tempStr);
    }

    tempStr = e.getAttribute(LINE_START_DECORATION_ATTR);
    if (notEmpty(tempStr)) {
        lineStartDecoration = LineDecorationType.valueOf(tempStr);
    }

    tempStr = e.getAttribute(LINE_END_DECORATION_ATTR);
    if (notEmpty(tempStr)) {
        lineEndDecoration = LineDecorationType.valueOf(tempStr);
    }

    tempStr = e.getAttribute(BOX_SHAPE_ATTR);
    if (notEmpty(tempStr)) {
        boxShapeType = BoxShapeType.valueOf(tempStr);
    }

    tempStr = e.getAttribute(PADDING_ATTR);
    if (notEmpty(tempStr)) {
        padding = Double.parseDouble(tempStr);
    }

    tempStr = e.getAttribute(SHADOW_ATTR);
    if (notEmpty(tempStr)) {
        shadow = Boolean.parseBoolean(tempStr);
    }

}

private boolean notEmpty(String tempStr) {
    if (tempStr == null)
        return false;
    if (tempStr.equals(""))
        return false;
    return true;
}

//@Override
public StyleImpl copy() {
    StyleImpl newStyle = new StyleImpl();
    return newStyle;
}

protected void updateOwner() {
    if (delayUpdate)
        return;
    if (owner != null) {
        owner.updateAll();
    }
}

@Override
public JMenu getMenu(StyleMenuType menuType) {


    JMenu menu = new JMenu("Style");

    if (menuType == StyleMenuType.LINE) {
        menu.add(getLineStartDecorationTypeSubmenu());
        menu.add(getLineEndDecorationTypeSubmenu());
    }
    menu.add(getStrokeColorMenuItem());
    menu.add(getStrokeWidthSubmenu());

    if (menuType == StyleMenuType.BOX) {
        menu.addSeparator();

        menu.add(getBoxShapeTypeSubmenu());
        menu.add(getFillColorMenuItem());
        menu.add(getGradientColorMenuItem());
        menu.add(getGradientTypeSubmenu());
    }

    if ((menuType == StyleMenuType.BOX) || (menuType == StyleMenuType.LINE)) {
        menu.addSeparator();

        menu.add(getFontSizeSubmenu());
        menu.add(getFontVariationsSubmenu());
        menu.add(getFontTypefaceSubmenu());
        menu.add(getTextColorMenuItem());
    }

    return menu;
}

@Override
public void delayUpdate(boolean inUpdate) {
    if (this.delayUpdate != inUpdate) {
        this.delayUpdate = inUpdate;
        if (!inUpdate)
            updateOwner();
    }
}

protected JMenuItem menuCommand(String text, String command, ActionListener listener) {
    JMenuItem item = new JMenuItem(text);
    item.setActionCommand(command);
    item.addActionListener(listener);
    return item;
}

protected JMenuItem menuCommand(String text, String command, ActionListener listener, boolean selected) {
    JMenuItem item = new JCheckBoxMenuItem(text);
    item.setActionCommand(command);
    item.addActionListener(listener);
    item.setSelected(selected);
    return item;
}

protected static class Commands {

    static final String EDIT_STYLE_FILLCOLOR = "Fill color...";
    static final String EDIT_STYLE_FILLCOLOR_COMMAND = "style-set-fill-color";
    static final String EDIT_STYLE_STROKECOLOR = "Line color...";
    static final String EDIT_STYLE_STROKECOLOR_COMMAND = "style-set-line-color";
    static final String EDIT_STYLE_TEXTCOLOR = "Text color...";
    static final String EDIT_STYLE_TEXTCOLOR_COMMAND = "style-set-text-color";
    static final String EDIT_STYLE_GRADIENTCOLOR = "Gradient color...";
    static final String EDIT_STYLE_GRADIENTCOLOR_COMMAND = "style-set-gradient-color";

}


}
