/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;

import lv.lumii.test.OpusFrame;
import lv.lumii.deek.util.Globals;

public class OpusOverview extends JDialog {

private DDrawing drawing;
private OpusFrame frame;
private OverviewPanel overviewPanel;

public OpusOverview(DDrawing drawing, OpusFrame mainFrame) {
    this(mainFrame);
    this.drawing = drawing;
    this.frame = mainFrame;

    initComponents();
}

public OpusOverview(OpusFrame mainFrame) {
    super(mainFrame);
}

public void initComponents() {
    overviewPanel = new OverviewPanel();
    add(overviewPanel, BorderLayout.CENTER);

    setSize(overviewPanel.getPreferredSize());
    setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
    setTitle("Overview");
    setAutoRequestFocus(false);
    setType(Type.UTILITY);

    addComponentListener(new ComponentListener() {

        @Override
        public void componentResized(ComponentEvent e) {
            updateView();
        }

        @Override
        public void componentMoved(ComponentEvent e) {
        }

        @Override
        public void componentShown(ComponentEvent e) {
            updateView();
        }

        @Override
        public void componentHidden(ComponentEvent e) {
        }

    });

}

public void updateView() {
    overviewPanel.recalculate();
    repaint();
}

public class OverviewPanel extends JComponent {

    private final int OVERVIEW_WIDTH = 200;
    private final int OVERVIEW_HEIGHT = OVERVIEW_WIDTH;
    private final Color PANE_FILL_COLOR = Color.DARK_GRAY;
    private final Color DIAGRAM_BACKGROUND_COLOR = Color.WHITE;
    private final Color DIAGRAM_BORDER_COLOR = Color.LIGHT_GRAY;
    private final Color POSITION_FILL_COLOR = new Color(0.5f, 0, 0, 0.2f);
    private final Color POSITION_BORDER_COLOR = new Color(1.0f, 0, 0, 1.0f);
    MouseHandler mh = new MouseHandler();
    private AffineTransform at = new AffineTransform();
    private Rectangle2D.Double drawingRect;
    private Rectangle2D.Double visibleRect;

    public OverviewPanel() {
        super();

        addMouseListener(mh);
        addMouseMotionListener(mh);
    }

    @Override
    public void paintComponent(Graphics g) {

        drawingRect = drawing.getBounds();
        if (drawingRect == null) return;

        visibleRect = frame.getViewport().zoomer.getVisibleArea();

        Graphics2D g2 = (Graphics2D) g.create();

        g2.setColor(PANE_FILL_COLOR);
        g2.fillRect(0, 0, getWidth(), getHeight());

        setTransformation(g2);

        g2.setColor(DIAGRAM_BACKGROUND_COLOR);
        g2.fill(drawingRect);

        drawing.draw(g2);

        g2.setColor(DIAGRAM_BORDER_COLOR);
        g2.draw(drawingRect);

        if (visibleRect.contains(drawingRect)) {

        } else {
            g2.setColor(POSITION_FILL_COLOR);
            g2.fill(visibleRect);
            g2.setColor(POSITION_BORDER_COLOR);
            g2.draw(visibleRect);
        }

        g2.dispose();
    }

    public void recalculate() {

        drawingRect = drawing.getBounds();
        if (drawingRect == null) {
            drawingRect = new Rectangle2D.Double(0, 0, 0, 0);
        }
        drawingRect.add(drawingRect.getMaxX() + Globals.WORLD_PADDING, drawingRect.getMaxY() + Globals.WORLD_PADDING);
        drawingRect.add(drawingRect.getMinX() - Globals.WORLD_PADDING, drawingRect.getMinY() - Globals.WORLD_PADDING);

        double scaleX = (double) this.getWidth() / drawingRect.width;
        double scaleY = (double) this.getHeight() / drawingRect.height;

        double scale = Math.min(scaleX, scaleY);
        double tx = -drawingRect.getMinX() * scale;
        double ty = -drawingRect.getMinY() * scale;

        at.setToIdentity();
        at.translate(tx, ty);
        at.scale(scale, scale);
    }

    public void setTransformation(Graphics2D g) {
        recalculate();
        g.setTransform(at);
    }

    @Override
    public Dimension getMinimumSize() {
        return new Dimension(OVERVIEW_WIDTH, OVERVIEW_HEIGHT);
    }

    @Override
    public Dimension getPreferredSize() {
        return getMinimumSize();
    }

    private boolean isPointVisible(Point p) {
        return false;
    }

    private class MouseHandler extends MouseAdapter {

        private Point prevPoint;
        private Point nextPoint;
        private boolean isMoveCursor = false;

        @Override
        public void mousePressed(MouseEvent e) {
            if (e.getButton() == MouseEvent.BUTTON1) {
                prevPoint = e.getPoint();
                nextPoint = prevPoint;
            }
        }

        @Override
        public void mouseDragged(MouseEvent e) {


            prevPoint = nextPoint;
            nextPoint = e.getPoint();

            if (!isMoveCursor) {
                setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
                isMoveCursor = true;
            }

            doPaning(nextPoint, prevPoint);
        }

        @Override
        public void mouseMoved(MouseEvent e) {

        }

        @Override
        public void mouseClicked(MouseEvent e) {

            if (e.getButton() == MouseEvent.BUTTON1) {
                doPaning(e.getPoint());
            }

        }

        @Override
        public void mouseReleased(MouseEvent e) {

            if (isMoveCursor) {
                setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                isMoveCursor = false;
            }

        }

        /**
         * Pane by the vector (p2 - p1)
         *
         * @param p2
         * @param p1
         */
        private void doPaning(Point p2, Point p1) {

            try {
                // TODO - optimize this
                AffineTransform ati = at.createInverse();

                Point2D.Double p1d = new Point2D.Double(p2.x, p2.y);
                Point2D.Double p2d = new Point2D.Double(p1.x, p1.y);

                Point2D p1world, p2world;
                p1world = ati.transform(p1d, null);
                p2world = ati.transform(p2d, null);

                frame.getViewport().zoomer.doPaning(
                        p1world.getX() - p2world.getX(),
                        p1world.getY() - p2world.getY()
                );

            } catch (NoninvertibleTransformException e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }

        }

        /**
         * Pane to the point
         *
         * @param p2 center this point in the window
         */
        private void doPaning(Point p2) {

            try {
                // TODO - optimize this
                AffineTransform ati = at.createInverse();

                Point2D.Double newCenterPoint = new Point2D.Double(p2.getX(), p2.getY());
                Point2D newCenterPointW = ati.transform(newCenterPoint, null);

                Point2D.Double oldCenterPointW = frame.getViewport().zoomer.getVisibleAreaCenter();

                frame.getViewport().zoomer.doPaning(
                        newCenterPointW.getX() - oldCenterPointW.getX(),
                        newCenterPointW.getY() - oldCenterPointW.getY()
                );

            } catch (NoninvertibleTransformException e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }


        }

    }

}
}
