/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.ui.tools;

import java.awt.Point;
import java.awt.event.MouseEvent;
import java.util.ArrayList;

import lv.lumii.deek.ui.DViewport;
import lv.lumii.deek.util.Globals;

/**
 * <p/>
 * Simplistic editor for ortholinear lines.
 * Corner dragging and segment dragging supported; number of segments does not change.
 * Click outside the line to finish editing.
 */
@SuppressWarnings("serial")
public class DReshapeOrthoLineTool extends DLineTool {

static final int CORNER_DISTANCE = 10;
static final int SEGMENT_DISTANCE = 5;
private HitTestResult hitTestResult = HitTestResult.OUTSIDE;
private int draggableCorner;
private boolean isCornerDragHV;
private int draggableSegment;

public DReshapeOrthoLineTool(DViewport owner, ArrayList<Point> line) {
    this.owner = owner;
    tempLine = line;
    initTool();
}

public DReshapeOrthoLineTool(DViewport owner, ArrayList<Point> line, MouseEvent e) {
    this(owner, line);
    mousePressed(e);
}

private void initTool() {
    drawCornerHandles = true;
}

private HitTestResult hitTest(Point p) {

    if (tempLine != null) {
        int nearestCorner = -1;
        int nearestDistance = Integer.MAX_VALUE;
        int thisDistance;

        // iterate corners, except ends
        for (int i = 1; i < tempLine.size() - 1; i++) {

            thisDistance = Math.abs(p.x - tempLine.get(i).x) +
                    Math.abs(p.y - tempLine.get(i).y);

            if (thisDistance > CORNER_DISTANCE) continue;

            if (thisDistance < nearestDistance) {
                nearestCorner = i;
                nearestDistance = thisDistance;
            }

        }
        if (nearestCorner != -1) {
            // corner hit
            draggableCorner = nearestCorner;
//				System.out.print("corner: ");
//				System.out.println(draggableCorner);
            return HitTestResult.CORNER;
        }

        int nearestSegment = -1;
        nearestDistance = Integer.MAX_VALUE;

        for (int i = 0; i < tempLine.size() - 1; i++) {

            Point c1 = tempLine.get(i);
            Point c2 = tempLine.get(i + 1);

            if (c1.x == c2.x) {
                if ((p.y >= Math.min(c1.y, c2.y)) && (p.y <= Math.max(c1.y, c2.y))) {
                    thisDistance = Math.abs(c1.x - p.x);
                } else {
                    thisDistance = Integer.MAX_VALUE;
                }
            } else {
                if ((p.x >= Math.min(c1.x, c2.x)) && (p.x <= Math.max(c1.x, c2.x))) {
                    thisDistance = Math.abs(c1.y - p.y);
                } else {
                    thisDistance = Integer.MAX_VALUE;
                }

            }

            if (thisDistance > SEGMENT_DISTANCE) continue;

            if (thisDistance < nearestDistance) {
                nearestSegment = i;
                nearestDistance = thisDistance;
            }
            if (nearestSegment != -1) {
                // segment hit
                draggableSegment = nearestSegment;
//					System.out.print("segment: ");
//					System.out.println(draggableSegment);
                return HitTestResult.SEGMENT;
            }

        }
    }

    return HitTestResult.OUTSIDE;
}

public void onPrimaryDown() {
    // hit test; if outside - end tool
    hitTestResult = hitTest(mouseDownPoint);

    if (hitTestResult == HitTestResult.OUTSIDE) {
        // return tempLine
        finish();
    }

    if (hitTestResult == HitTestResult.CORNER) {
        if (tempLine.get(draggableCorner - 1).x == tempLine.get(draggableCorner).x) {
            isCornerDragHV = false;
        } else {
            isCornerDragHV = true;
        }
    }
}

public void onDrag() {
    // do corner adjustments to the tempLine

    Point prevCorner, thisCorner, nextCorner;

    switch (hitTestResult) {

        case CORNER:
            prevCorner = tempLine.get(draggableCorner - 1);
            thisCorner = tempLine.get(draggableCorner);
            nextCorner = tempLine.get(draggableCorner + 1);

            if (isCornerDragHV) {
                prevCorner.y += dragVector.height;
                nextCorner.x += dragVector.width;
            } else {
                prevCorner.x += dragVector.width;
                nextCorner.y += dragVector.height;
            }

            thisCorner.x += dragVector.width;
            thisCorner.y += dragVector.height;

//			if (prevCorner.x == thisCorner.x) {
//				// prev segment vertical
//				prevCorner.x += dragVector.width;
//			} else {
//				// prev segment horizontal
//				prevCorner.y += dragVector.height;
//			}
//
//			if (nextCorner.x == thisCorner.x) {
//				// next segment vertical
//				nextCorner.x += dragVector.width;
//			} else {
//				// next segment horizontal
//				nextCorner.y += dragVector.height;
//			}
//
//			thisCorner.x += dragVector.width;
//			thisCorner.y += dragVector.height;

            // TODO - optimized repaint
            repaint();
            break;

        case SEGMENT:
            prevCorner = tempLine.get(draggableSegment);
            nextCorner = tempLine.get(draggableSegment + 1);

            if (prevCorner.x == nextCorner.x) {
                // vertical segment
                prevCorner.x += dragVector.width;
                nextCorner.x += dragVector.width;
            } else {
                // horizontal segment
                prevCorner.y += dragVector.height;
                nextCorner.y += dragVector.height;
            }
            // TODO - optimized repaint
            repaint();
            break;

        default:
            // error

    }

}

public void onPrimaryUp() {

    // nothing if in dev mode

    if (!Globals.isDeveloperMode())
        finish();

}

public void finish() {
    owner.endReshapeOrthoLine(tempLine);
}

private enum HitTestResult {OUTSIDE, CORNER, SEGMENT}

}
