/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.ui.tools;

import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.util.ArrayList;

import lv.lumii.deek.util.Globals;
import lv.lumii.deek.ui.DViewport;

/**
 * <p/>
 * Simplistic editor for straight (and poly?) lines.
 * Drag line (line segments), or drag end points (and corners).
 * Click outside the line to finish editing.
 */
@SuppressWarnings("serial")
public class DReshapePolyLineTool extends DLineTool {

static final int CORNER_DISTANCE = 10;
static final int SEGMENT_DISTANCE = 5;
private HitTestResult hitTestResult = HitTestResult.OUTSIDE;
private int draggableCorner;
private int draggableSegment;


public DReshapePolyLineTool(DViewport owner, ArrayList<Point> line) {
    this.owner = owner;
    tempLine = line;
    initTool();
}

public DReshapePolyLineTool(DViewport owner, ArrayList<Point> line, MouseEvent e) {
    this(owner, line);
    mousePressed(e);
}

private void initTool() {
    drawCornerHandles = true;
}

private HitTestResult hitTest(Point p) {

    if (tempLine != null) {

        int nearestCorner = -1;
        int nearestDistance = Integer.MAX_VALUE;
        int thisDistance;

        // iterate corners, including ends
        for (int i = 0; i <= tempLine.size() - 1; i++) {

            thisDistance = (int) Math.sqrt(
                    (p.x - tempLine.get(i).x) * (p.x - tempLine.get(i).x) +
                            (p.y - tempLine.get(i).y) * (p.y - tempLine.get(i).y));

            if (thisDistance > CORNER_DISTANCE) continue;

            if (thisDistance < nearestDistance) {
                nearestCorner = i;
                nearestDistance = thisDistance;
            }

        }
        if (nearestCorner != -1) {
            // corner hit
            draggableCorner = nearestCorner;
            return HitTestResult.CORNER;
        }

        int nearestSegment = -1;
        nearestDistance = Integer.MAX_VALUE;

        for (int i = 0; i < tempLine.size() - 1; i++) {

            Point c1 = tempLine.get(i);
            Point c2 = tempLine.get(i + 1);

            thisDistance = (int) Line2D.ptSegDist(c1.x, c1.y, c2.x, c2.y, p.x, p.y);

            if (thisDistance > SEGMENT_DISTANCE) continue;

            if (thisDistance < nearestDistance) {
                nearestSegment = i;
                nearestDistance = thisDistance;
            }
        }

        if (nearestSegment != -1) {
            // segment hit

            if (!ctrlPressed) {

                draggableSegment = nearestSegment;
                return HitTestResult.SEGMENT;

            } else {

                nearestCorner = nearestSegment + 1;
                tempLine.add(nearestCorner, new Point(p));
                draggableCorner = nearestCorner;
                return HitTestResult.CORNER;

            }
        }

    }

    return HitTestResult.OUTSIDE;
}

public void onPrimaryDown() {
    // hit test; if outside - end tool
    hitTestResult = hitTest(mouseDownPoint);

    if (hitTestResult == HitTestResult.OUTSIDE) {
        finish();
    }

}

public void onDrag() {
    // do corner adjustments to the tempLine

    Point prevCorner, thisCorner, nextCorner;

    switch (hitTestResult) {

        case CORNER:
            thisCorner = tempLine.get(draggableCorner);

            thisCorner.x += dragVector.width;
            thisCorner.y += dragVector.height;

            // TODO - optimized repaint
            repaint();
            break;

        case SEGMENT:
            prevCorner = tempLine.get(draggableSegment);
            nextCorner = tempLine.get(draggableSegment + 1);

            prevCorner.x += dragVector.width;
            nextCorner.x += dragVector.width;
            prevCorner.y += dragVector.height;
            nextCorner.y += dragVector.height;

            // TODO - optimized repaint
            repaint();
            break;

        default:
            // error

    }

}

@Override
public void onPrimaryUp() {

    // on ctrl-click remove the corner
    if ((!mouseHasBeenDragged) && (hitTestResult == HitTestResult.CORNER) && (ctrlPressed)) {
        if ((draggableCorner > 0) && (draggableCorner < tempLine.size() - 1)) {
            tempLine.remove(draggableCorner);
            repaint();
        }
    }

    if (!Globals.isDeveloperMode())
        finish();

}

@Override
public void finish() {
    owner.endReshapePolyLine(tempLine); // FIXME
}

private enum HitTestResult {OUTSIDE, CORNER, SEGMENT}

}
