/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.util;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.util.Arrays;
import java.util.List;

import lv.lumii.deek.style.LineDecorationType;
import lv.lumii.deek.util.Globals;

/**
 * Helper class for draw operations.
 */
public abstract class Drawer {

private static final int MAXPOINTS = 4;
private static boolean localDump = false; // for debugging
private static boolean oldDraw = false; // true: int, false: float
private static Globals globals = Globals.getInstance();
// ----------------------------------------------------------
private static double v21x, v21y, v21x90, v21y90, v21x270, v21y270;
private static double[] dPointsX = new double[MAXPOINTS];

// ----------------------------------------------------------
private static double[] dPointsY = new double[MAXPOINTS];
private static int[] iPointsX = new int[MAXPOINTS];

// ----------------------------------------------------------
private static int[] iPointsY = new int[MAXPOINTS];
private static int numPoints;

// ----------------------------------------------------------

private static OrthoSegmentDirection getOrthoSegmentDirection(Point2D p1,
                                                              Point2D p2) {
    return getOrthoSegmentDirection(p1.getX(), p1.getY(), p2.getX(),
            p2.getY());
}

// ----------------------------------------------------------

private static OrthoSegmentDirection getOrthoSegmentDirection(double x1,
                                                              double y1, double x2, double y2) {
    if (x1 == x2) {
        if (y2 > y1)
            return OrthoSegmentDirection.D270;
        else
            return OrthoSegmentDirection.D90;
    } else if (x2 > x1)
        return OrthoSegmentDirection.D0;
    else
        return OrthoSegmentDirection.D180;
}

/**
 * Draws a decorated straight line between the given points
 *
 * @param g          graphic context
 * @param points     line points
 * @param startType  decoration type at line start
 * @param finishType decoration type at line end
 */
public static void drawDecoratedStraightLine(Graphics2D g, List<Point2D.Double> points,
                                             LineDecorationType startType, LineDecorationType finishType) {

    Point2D prevPoint, nextPoint;

    prevPoint = points.get(0);
    nextPoint = points.get(points.size() - 1);

    if (localDump) {
        System.out.println(String.format("draw straight line: from %s to %s", prevPoint.toString(), nextPoint.toString()));
    }

    if (oldDraw) {
        g.drawLine((int) prevPoint.getX(), (int) prevPoint.getY(),
                (int) nextPoint.getX(), (int) nextPoint.getY());
    } else {
        Path2D.Double path = new Path2D.Double();
        path.moveTo(prevPoint.getX(), prevPoint.getY());
        path.lineTo(nextPoint.getX(), nextPoint.getY());
        g.draw(path);
    }

    if (startType != LineDecorationType.NONE) {
        drawLineEndDecoration(g,
                nextPoint.getX(), nextPoint.getY(),
                prevPoint.getX(), prevPoint.getY(),
                startType);
    }

    if (finishType != LineDecorationType.NONE) {
        drawLineEndDecoration(g,
                prevPoint.getX(), prevPoint.getY(),
                nextPoint.getX(), nextPoint.getY(),
//					LineEndType.FILLED_DIAMOND);
                finishType);
    }
}

/**
 * Draws a straight line between the given points
 *
 * @param g      graphic context
 * @param points line points
 */
public static void drawDecoratedStraightLine(Graphics2D g, List<Point2D.Double> points) {
    drawDecoratedStraightLine(g, points, LineDecorationType.NONE, LineDecorationType.NONE);
}

/**
 * Draws a decorated poly line between the given points
 *
 * @param g          graphic context
 * @param points     line points
 * @param startType  decoration type at line start
 * @param finishType decoration type at line end
 */
public static void drawDecoratedPolyLine(Graphics2D g, List<Point2D.Double> points,
                                         LineDecorationType startType, LineDecorationType finishType) {

    Point2D prevPoint, nextPoint;
    Path2D path;

    if (points.size() > 1) {

        prevPoint = points.get(0);

        if (oldDraw) {

            for (int i = 1; i < points.size(); i++) {
                nextPoint = points.get(i);
                g.drawLine((int) prevPoint.getX(), (int) prevPoint.getY(),
                        (int) nextPoint.getX(), (int) nextPoint.getY());
                prevPoint = nextPoint;
            }

        } else {
            path = new Path2D.Double();
            path.moveTo(prevPoint.getX(), prevPoint.getY());

            for (int i = 1; i < points.size(); i++) {
                nextPoint = points.get(i);

                path.lineTo(nextPoint.getX(), nextPoint.getY());

                prevPoint = nextPoint;
            }

            g.draw(path);

        }

        if (startType != LineDecorationType.NONE) {
            prevPoint = points.get(0);
            nextPoint = points.get(1);
            drawLineEndDecoration(g,
                    nextPoint.getX(), nextPoint.getY(),
                    prevPoint.getX(), prevPoint.getY(),
                    startType);
        }

        if (finishType != LineDecorationType.NONE) {
            prevPoint = points.get(points.size() - 2);
            nextPoint = points.get(points.size() - 1);
            drawLineEndDecoration(g,
                    prevPoint.getX(), prevPoint.getY(),
                    nextPoint.getX(), nextPoint.getY(),
                    finishType);
        }
    }

}

/**
 * Draws a poly line between the given points
 *
 * @param g      graphic context
 * @param points line points
 */
public static void drawDecoratedPolyLine(Graphics2D g, List<Point2D.Double> points) {
    drawDecoratedPolyLine(g, points, LineDecorationType.NONE, LineDecorationType.NONE);
}

/**
 * Draws a decorated ortho line between the given points
 *
 * @param g          graphic context
 * @param points     line points
 * @param startType  decoration type at line start
 * @param finishType decoration type at line end
 */
public static void drawDecoratedOrthoLine(Graphics2D g, List<Point2D.Double> points,
                                          LineDecorationType startType, LineDecorationType finishType) {

    Point2D prevPoint, nextPoint;
    OrthoSegmentDirection prevDirection, thisDirection, nextDirection;

    if (points.size() > 1) {

        if (oldDraw) {

            prevPoint = points.get(0);
            for (int i = 1; i < points.size(); i++) {
                nextPoint = points.get(i);

                if (i == 1)
                    prevDirection = OrthoSegmentDirection.NONE;
                else {
                    prevDirection = getOrthoSegmentDirection(points.get(i - 2),
                            points.get(i - 1));
                }
                thisDirection = getOrthoSegmentDirection(points.get(i - 1),
                        points.get(i));
                if (i == points.size() - 1)
                    nextDirection = OrthoSegmentDirection.NONE;
                else
                    nextDirection = getOrthoSegmentDirection(points.get(i),
                            points.get(i + 1));

                // g.drawLine((int)prevPoint.getX(), (int)prevPoint.getY(),
                // (int)nextPoint.getX(), (int)nextPoint.getY());
                drawDecoratedOrthoSegment(g, prevPoint, nextPoint,
                        prevDirection, thisDirection, nextDirection,
                        startType, finishType);

                prevPoint = nextPoint;
            }

        } else {
            // TODO
            g.draw(buildOrthoLinePath2D(points));

            prevPoint = points.get(1);
            nextPoint = points.get(0);
            drawOrthoLineEndDecoration2D(g, prevPoint.getX(), prevPoint.getY(),
                    nextPoint.getX(), nextPoint.getY(),
                    startType);

            prevPoint = points.get(points.size() - 2);
            nextPoint = points.get(points.size() - 1);
            drawOrthoLineEndDecoration2D(g, prevPoint.getX(), prevPoint.getY(),
                    nextPoint.getX(), nextPoint.getY(),
                    finishType);

        }
    }

}

/**
 * Draws an ortho line between the given points
 *
 * @param g      graphic context
 * @param points line points
 */

public static void drawDecoratedOrthoLine(Graphics2D g, List<Point2D.Double> points) {
    drawDecoratedOrthoLine(g, points, LineDecorationType.NONE, LineDecorationType.NONE);
}

/**
 * Draws a decorated spline line between the given points
 *
 * @param g          graphic context
 * @param points     line points
 * @param startType  decoration type at line start
 * @param finishType decoration type at line end
 */
public static void drawDecoratedSpline(Graphics2D g, List<Point2D.Double> points,
                                       LineDecorationType startType, LineDecorationType finishType) {

    Point2D prevPoint, nextPoint;
    double prevHalfPointX, prevHalfPointY, nextHalfPointX, nextHalfPointY;
    Path2D path;

    if (points.size() < 2) return;

    if (points.size() > 2) {
        prevPoint = points.get(0);

        path = new Path2D.Double();
        path.moveTo(prevPoint.getX(), prevPoint.getY());

        for (int i = 1; i < points.size() - 1; i++) {
            nextPoint = points.get(i);
            prevHalfPointX = (points.get(i - 1).getX() + points.get(i).getX()) / 2;
            prevHalfPointY = (points.get(i - 1).getY() + points.get(i).getY()) / 2;
            nextHalfPointX = (points.get(i).getX() + points.get(i + 1).getX()) / 2;
            nextHalfPointY = (points.get(i).getY() + points.get(i + 1).getY()) / 2;

            if (i == 1) {
                path.lineTo(prevHalfPointX, prevHalfPointY);
            }

            path.curveTo(
                    nextPoint.getX(), nextPoint.getY(),
                    nextPoint.getX(), nextPoint.getY(),
                    nextHalfPointX, nextHalfPointY
            );

            if (i == points.size() - 2) {
                path.lineTo(points.get(points.size() - 1).getX(), points.get(points.size() - 1).getY());
            }

            prevPoint = nextPoint;
        }
        g.draw(path);


        if (startType != LineDecorationType.NONE) {
            prevPoint = points.get(0);
            nextPoint = points.get(1);
            drawLineEndDecoration(g,
                    nextPoint.getX(), nextPoint.getY(),
                    prevPoint.getX(), prevPoint.getY(),
                    startType);
        }

        if (finishType != LineDecorationType.NONE) {
            prevPoint = points.get(points.size() - 2);
            nextPoint = points.get(points.size() - 1);
            drawLineEndDecoration(g,
                    prevPoint.getX(), prevPoint.getY(),
                    nextPoint.getX(), nextPoint.getY(),
                    finishType);
        }
    } else // 2 points
    {
        // TODO
        prevPoint = points.get(0);

        path = new Path2D.Double();

        path.moveTo(points.get(0).getX(), points.get(0).getY());
        path.lineTo(points.get(1).getX(), points.get(1).getY());

        g.draw(path);

        if (startType != LineDecorationType.NONE) {
            prevPoint = points.get(0);
            nextPoint = points.get(1);
            drawLineEndDecoration(g,
                    nextPoint.getX(), nextPoint.getY(),
                    prevPoint.getX(), prevPoint.getY(),
                    startType);
        }

        if (finishType != LineDecorationType.NONE) {
            prevPoint = points.get(points.size() - 2);
            nextPoint = points.get(points.size() - 1);
            drawLineEndDecoration(g,
                    prevPoint.getX(), prevPoint.getY(),
                    nextPoint.getX(), nextPoint.getY(),
                    finishType);
        }

    }
}

// ----------------------------------------------------------

/**
 * Draws a spline line between the given points
 *
 * @param g      graphic context
 * @param points line points
 */

public static void drawDecoratedSpline(Graphics2D g, List<Point2D.Double> points) {
    drawDecoratedSpline(g, points, LineDecorationType.NONE, LineDecorationType.NONE);
}

private static void drawOrthoLineStartDecoration(Graphics2D g, double x1, double y1,
                                                 double x2, double y2, OrthoSegmentDirection thisDirection,
                                                 LineDecorationType lineEndType) {

    final int AW = 4;
    final int AL = 10;

    final int DW = 2;
    final int DL = 4;

    final int CR = 2;
    final int CR2 = 2 * CR;

    switch (lineEndType) {
        case NONE:
            break;

        case FILLED_CIRCLE:
            g.fillOval((int) x1 - CR, (int) y1 - CR, CR2, CR2);
            break;

        case FILLED_ARROW:
        case HOLLOW_ARROW:
        case STANDARD_ARROW:
            int[] xPoints = {(int) x1, (int) x1, (int) x1};
            int[] yPoints = {(int) y1, (int) y1, (int) y1};
            switch (thisDirection) {
                case D180:
                    xPoints[0] -= AL;
                    yPoints[0] -= AW;
                    xPoints[2] -= AL;
                    yPoints[2] += AW;
                    break;

                case D0:
                    xPoints[0] += AL;
                    yPoints[0] -= AW;
                    xPoints[2] += AL;
                    yPoints[2] += AW;
                    break;

                case D270:
                    xPoints[0] -= AW;
                    yPoints[0] += AL;
                    xPoints[2] += AW;
                    yPoints[2] += AL;
                    break;

                case D90:
                    xPoints[0] -= AW;
                    yPoints[0] -= AL;
                    xPoints[2] += AW;
                    yPoints[2] -= AL;
                    break;

            }
            if (lineEndType == LineDecorationType.FILLED_ARROW)
                g.fillPolygon(xPoints, yPoints, xPoints.length);
            else if (lineEndType == LineDecorationType.HOLLOW_ARROW)
                g.drawPolygon(xPoints, yPoints, xPoints.length);
            else // STANDARD_ARROW
                g.drawPolyline(xPoints, yPoints, xPoints.length);

            break;

        case FILLED_DIAMOND:
            int[] xPoints4 = {(int) x1, (int) x1, (int) x1, (int) x1};
            int[] yPoints4 = {(int) y1, (int) y1, (int) y1, (int) y1};

            switch (thisDirection) {
                case D180:
                    xPoints4[1] -= DL;
                    yPoints4[1] -= DW;
                    xPoints4[2] -= 2 * DL;
                    xPoints4[3] -= DL;
                    yPoints4[3] += DW;
                    break;

                case D0:
                    xPoints4[1] += DL;
                    yPoints4[1] -= DW;
                    xPoints4[2] += 2 * DL;
                    xPoints4[3] += DL;
                    yPoints4[3] += DW;
                    break;

                case D270:
                    xPoints4[1] -= DW;
                    yPoints4[1] += DL;
                    yPoints4[2] += 2 * DL;
                    xPoints4[3] += DW;
                    yPoints4[3] += DL;
                    break;

                case D90:
                    xPoints4[1] -= DW;
                    yPoints4[1] -= DL;
                    yPoints4[2] -= 2 * DL;
                    xPoints4[3] += DW;
                    yPoints4[3] -= DL;
                    break;

            }
            g.fillPolygon(xPoints4, yPoints4, xPoints4.length);

    }

}

private static Path2D buildOrthoLinePath2D(List<Point2D.Double> points) {

    Point2D p1, p2;

    double x1, y1, x2, y2;

    double xa, ya;

    double r = globals.lineCornerRounding;
    double rr = 2 * r;

    double sa, aa;

    OrthoSegmentDirection prevDirection, thisDirection, nextDirection;

    Path2D.Double path = new Path2D.Double();

    if (points.size() > 1) {

        p1 = points.get(0);
        x1 = p1.getX();
        y1 = p1.getY();

        path.moveTo(x1, y1);

        for (int i = 1; i < points.size(); i++) {

            p2 = points.get(i);
            x2 = p2.getX();
            y2 = p2.getY();

            if (i == 1)
                prevDirection = OrthoSegmentDirection.NONE;
            else {
                prevDirection = getOrthoSegmentDirection(points.get(i - 2),
                        points.get(i - 1));
            }

            thisDirection = getOrthoSegmentDirection(points.get(i - 1),
                    points.get(i));
            if (i == points.size() - 1)
                nextDirection = OrthoSegmentDirection.NONE;
            else
                nextDirection = getOrthoSegmentDirection(points.get(i),
                        points.get(i + 1));

            if (nextDirection != OrthoSegmentDirection.NONE) {
                // calculate points for segment (x2, y2) and arc (xa, ya, wa, ha, startAngle, arcAngle)
                // TODO

                xa = x2;
                ya = y2;
                sa = 0;
                aa = 0;

                switch (thisDirection) {
                    case D0:
                        xa = x2 - rr;
                        x2 -= r;
                        switch (nextDirection) {
                            case D90:
                                ya = y2 - rr;
                                sa = -90;
                                aa = 90;
                                break;

                            case D270:
                                ya = y2;
                                sa = 90;
                                aa = -90;
                                break;

                            default:
                                break;
                        }
                        break;

                    case D180:
                        xa = x2;
                        x2 += r;
                        switch (nextDirection) {
                            case D90:
                                ya = y2 - rr;
                                sa = -90;
                                aa = -90;
                                break;

                            case D270:
                                ya = y2;
                                sa = 90;
                                aa = 90;
                                break;

                            default:
                                break;
                        }
                        break;

                    case D270:
                        ya = y2 - rr;
                        y2 -= r;
                        switch (nextDirection) {
                            case D0:
                                xa = x2;
                                sa = 180;
                                aa = 90;
                                break;
                            case D180:
                                xa = x2 - rr;
                                sa = 0;
                                aa = -90;
                                break;
                            default:
                                break;
                        }
                        break;

                    case D90:
                        ya = y2;
                        y2 += r;
                        switch (nextDirection) {
                            case D0:
                                xa = x2;
                                sa = 180;
                                aa = -90;
                                break;
                            case D180:
                                xa = x2 - rr;
                                sa = 0;
                                aa = 90;
                                break;
                            default:
                                break;
                        }
                        break;

                    default:
                        break;
                }

                // draw segment
                path.lineTo(x2, y2);
                // draw arc
                // TODO
                path.append(new Arc2D.Double(xa, ya, rr, rr, sa, aa, Arc2D.OPEN), true);

            } else {
                // draw segment
                path.lineTo(x2, y2);
            }


            p1 = p2;
        }
        return path;
    }

    return null;
}

private static int calculateArrowHead0() {

    final double AW = 4.0;
    final double AL = 10.0;

    shiftDPoint(0, v21x * AL + v21x90 * AW, v21y * AL + v21y90 * AW);
    shiftDPoint(2, v21x * AL + v21x270 * AW, v21y * AL + v21y270 * AW);

    return 3;
}

private static int calculateDiamond0() {

    final double AW = 2.0;
    final double AL = 4.0;
    final double AA = 1.0;

    shiftDPoint(0, v21x * AA, v21y * AA); // FIXME

    shiftDPoint(1, v21x * (AA + AL) + v21x270 * AW, v21y * (AA + AL) + v21y270 * AW);
    shiftDPoint(3, v21x * (AA + AL) - v21x270 * AW, v21y * (AA + AL) - v21y270 * AW);
    shiftDPoint(2, v21x * (AA + AL * 2), v21y * (AA + AL * 2));

    return 4;
}

private static void drawOrthoLineEndDecoration2D(Graphics2D g,
                                                 double x1, double y1,
                                                 double x2, double y2,
                                                 LineDecorationType lineEndType) {

    final double CR = 2;
    final double CR2 = 2 * CR;

    Path2D path;

    switch (lineEndType) {
        case NONE:
            break;

        case FILLED_CIRCLE:
            g.fill(new Ellipse2D.Double(x2 - CR, y2 - CR, CR2, CR2));
            break;

        case FILLED_ARROW:
        case HOLLOW_ARROW:
        case STANDARD_ARROW:
            calculateVectors21(x1, y1, x2, y2);
            setDPointsFromPoint(x2, y2);

            numPoints = calculateArrowHead0();

            path = new Path2D.Double();
            path.moveTo(dPointsX[0], dPointsY[0]);

            for (int i = 1; i < numPoints; i++) {
                path.lineTo(dPointsX[i], dPointsY[i]);
            }

            if (lineEndType == LineDecorationType.FILLED_ARROW) {
                path.closePath();
                g.fill(path);

            } else if (lineEndType == LineDecorationType.HOLLOW_ARROW) {
                path.closePath();
                g.draw(path);

            } else if (lineEndType == LineDecorationType.STANDARD_ARROW) {
                g.draw(path);

            }

            break;

        case FILLED_DIAMOND:
            calculateVectors21(x1, y1, x2, y2);
            setDPointsFromPoint(x2, y2);

            numPoints = calculateDiamond0();

            path = new Path2D.Double();
            path.moveTo(dPointsX[0], dPointsY[0]);

            for (int i = 1; i < numPoints; i++) {
                path.lineTo(dPointsX[i], dPointsY[i]);
            }

            path.closePath();
            g.fill(path);

    }
}

private static void drawOrthoLineFinishDecoration(Graphics2D g, double x1, double y1,
                                                  double x2, double y2, OrthoSegmentDirection thisDirection,
                                                  LineDecorationType lineEndType) {

    final int AW = 4;
    final int AL = 10;

    final int DW = 2;
    final int DL = 4;

    final int CR = 2;
    final int CR2 = 2 * CR;

    switch (lineEndType) {
        case NONE:
            break;

        case FILLED_CIRCLE:
            g.fillOval((int) x2 - CR, (int) y2 - CR, CR2, CR2);
            // g.drawOval((int)x2 - CR, (int)y2 - CR, CR2, CR2);
            break;

        case FILLED_ARROW:
        case HOLLOW_ARROW:
        case STANDARD_ARROW:
            int[] xPoints = {(int) x2, (int) x2, (int) x2};
            int[] yPoints = {(int) y2, (int) y2, (int) y2};
            switch (thisDirection) {
                case D0:
                    xPoints[0] -= AL;
                    yPoints[0] -= AW;
                    xPoints[2] -= AL;
                    yPoints[2] += AW;
                    break;

                case D180:
                    xPoints[0] += AL;
                    yPoints[0] -= AW;
                    xPoints[2] += AL;
                    yPoints[2] += AW;
                    break;

                case D90:
                    xPoints[0] -= AW;
                    yPoints[0] += AL;
                    xPoints[2] += AW;
                    yPoints[2] += AL;
                    break;

                case D270:
                    xPoints[0] -= AW;
                    yPoints[0] -= AL;
                    xPoints[2] += AW;
                    yPoints[2] -= AL;
                    break;

            }

            if (lineEndType == LineDecorationType.FILLED_ARROW)
                g.fillPolygon(xPoints, yPoints, xPoints.length);
            else if (lineEndType == LineDecorationType.HOLLOW_ARROW)
                g.drawPolygon(xPoints, yPoints, xPoints.length);
            else if (lineEndType == LineDecorationType.STANDARD_ARROW) {
                g.drawPolyline(xPoints, yPoints, xPoints.length);
            }

            break;

        case FILLED_DIAMOND:
            int[] xPoints4 = {(int) x2, (int) x2, (int) x2, (int) x2};
            int[] yPoints4 = {(int) y2, (int) y2, (int) y2, (int) y2};

            switch (thisDirection) {
                case D0:
                    xPoints4[1] -= DL;
                    yPoints4[1] -= DW;
                    xPoints4[2] -= 2 * DL;
                    xPoints4[3] -= DL;
                    yPoints4[3] += DW;
                    break;

                case D180:
                    xPoints4[1] += DL;
                    yPoints4[1] -= DW;
                    xPoints4[2] += 2 * DL;
                    xPoints4[3] += DL;
                    yPoints4[3] += DW;
                    break;

                case D90:
                    xPoints4[1] -= DW;
                    yPoints4[1] += DL;
                    yPoints4[2] += 2 * DL;
                    xPoints4[3] += DW;
                    yPoints4[3] += DL;
                    break;

                case D270:
                    xPoints4[1] -= DW;
                    yPoints4[1] -= DL;
                    yPoints4[2] -= 2 * DL;
                    xPoints4[3] += DW;
                    yPoints4[3] -= DL;
                    break;

            }
            g.fillPolygon(xPoints4, yPoints4, xPoints4.length);

    }

}

private static void drawDecoratedOrthoSegment(Graphics2D g, Point2D p1,
                                              Point2D p2, OrthoSegmentDirection prevDirection,
                                              OrthoSegmentDirection thisDirection,
                                              OrthoSegmentDirection nextDirection,
                                              LineDecorationType startType,
                                              LineDecorationType finishType) {

    double x1, y1, x2, y2;

    x1 = p1.getX();
    y1 = p1.getY();

    x2 = p2.getX();
    y2 = p2.getY();

    if (prevDirection != OrthoSegmentDirection.NONE) {
        switch (thisDirection) {
            case D0:
                x1 += globals.lineCornerRounding;
                break;
            case D90:
                y1 -= globals.lineCornerRounding;
                break;
            case D180:
                x1 -= globals.lineCornerRounding;
                break;
            case D270:
                y1 += globals.lineCornerRounding;
        }
    } else {
        drawOrthoLineStartDecoration(g, x1, y1, x2, y2, thisDirection,
                startType);

    }

    double xa, ya;
    double wa, ha;
    xa = x2;
    ya = y2;
    wa = ha = 2 * globals.lineCornerRounding;
    int startAngle, arcAngle;
    startAngle = arcAngle = 0;

    if (nextDirection != OrthoSegmentDirection.NONE) {
        switch (thisDirection) {
            case D0:
                xa = x2 - wa;
                x2 -= globals.lineCornerRounding;
                switch (nextDirection) {
                    case D90:
                        ya = y2 - ha;
                        startAngle = -90;
                        arcAngle = 90;
                        break;
                    case D270:
                        ya = y2;
                        startAngle = 90;
                        arcAngle = -90;
                        break;
                    default:
                        break;
                }
                break;

            case D180:
                xa = x2;
                x2 += globals.lineCornerRounding;
                switch (nextDirection) {
                    case D90:
                        ya = y2 - ha;
                        startAngle = 270;
                        arcAngle = -90;
                        break;
                    case D270:
                        ya = y2;
                        startAngle = 90;
                        arcAngle = 90;
                        break;
                    default:
                        break;
                }
                break;

            case D90:
                ya = y2;
                y2 += globals.lineCornerRounding;
                switch (nextDirection) {
                    case D0:
                        xa = x2;
                        startAngle = 180;
                        arcAngle = -90;
                        break;
                    case D180:
                        xa = x2 - wa;
                        startAngle = 0;
                        arcAngle = 90;
                        break;
                    default:
                        break;
                }
                break;

            case D270:
                ya = y2 - ha;
                y2 -= globals.lineCornerRounding;
                switch (nextDirection) {
                    case D0:
                        xa = x2;
                        startAngle = 180;
                        arcAngle = 90;
                        break;
                    case D180:
                        xa = x2 - wa;
                        startAngle = 0;
                        arcAngle = -90;
                        break;
                    default:
                        break;
                }
                break;

            default:
                break;

        }

        g.drawLine((int) x1, (int) y1, (int) x2, (int) y2);
        g.drawArc((int) xa, (int) ya, (int) wa, (int) ha, startAngle,
                arcAngle);

    } else {


        double x2f = x2;
        double y2f = y2;

        switch (thisDirection) {
            case D0:
                x2f = x2 - 3;
                break;
            case D180:
                x2f = x2 + 3;
                break;
            case D270:
                y2f = y2 - 3;
                break;
            case D90:
                y2f = y2 + 3;
                break;

        }
        g.drawLine((int) x1, (int) y1, (int) x2f, (int) y2f);
        drawOrthoLineFinishDecoration(g, x1, y1, x2, y2, thisDirection,
                finishType);

    }

}

private static void drawGlyph2D(Graphics2D g, int n, GlyphMode glyphMode) {

    Path2D.Double path = new Path2D.Double();

    if (localDump) {
        System.out.println(Arrays.toString(dPointsX));
        System.out.println(Arrays.toString(dPointsY));

        for (int k = 1; k < n; k++) {
            double dx = dPointsX[k] - dPointsX[k - 1];
            double dy = dPointsY[k] - dPointsY[k - 1];
            double d = Math.sqrt(dx * dx + dy * dy);
            System.out.println(String.format("%d: from [%f, %f] to [%f, %f] (dist=%f; dx=%f, dy=%f)", k, dPointsX[k - 1], dPointsY[k - 1], dPointsX[k], dPointsY[k], d, dx, dy));
        }
    }

    path.moveTo(dPointsX[0], dPointsY[0]);
    for (int i = 1; i < n; i++) {
        path.lineTo(dPointsX[i], dPointsY[i]);
    }

    switch (glyphMode) {
        case FILL:
            path.closePath();
            g.fill(path);
            g.draw(path);
            break;

        case HOLLOW:
            path.closePath();
            g.draw(path);
            break;

        case STROKE:
            g.draw(path);
            break;

    }
}

private static void drawGlyph(Graphics2D g, int n, GlyphMode glyphMode) {

    Graphics2D tempG2;

    switch (glyphMode) {
        case WHITE_FILL:
            tempG2 = (Graphics2D) g.create();
            tempG2.setColor(Color.WHITE);
            g.fillPolygon(iPointsX, iPointsY, n);
            tempG2.dispose();
            g.drawPolygon(iPointsX, iPointsY, n);

            break;

        case FILL:
            g.fillPolygon(iPointsX, iPointsY, n);
//			break;

        case HOLLOW:
            g.drawPolygon(iPointsX, iPointsY, n);

            break;

        case STROKE:
            g.drawPolyline(iPointsX, iPointsY, n);
            break;
    }

}

private static void calculateVectors21(double x1, double y1, double x2, double y2) {

    double dx = x2 - x1;
    double dy = y2 - y1;

    double len = Math.sqrt(dx * dx + dy * dy);

    v21x = -dx / len;
    v21y = -dy / len;

    v21x90 = -v21y;
    v21y90 = v21x;

    v21x270 = v21y;
    v21y270 = -v21x;
}

private static void setDPointsFromPoint(double x, double y) {

    for (int i = 0; i < MAXPOINTS; i++) {
        dPointsX[i] = x;
        dPointsY[i] = y;
    }

}

private static void setIPointsFromDPoints() {

    for (int i = 0; i < MAXPOINTS; i++) {
        iPointsX[i] = (int) dPointsX[i];
        iPointsY[i] = (int) dPointsY[i];
    }

}

private static void shiftDPoint(int num, double dx, double dy) {
    dPointsX[num] += dx;
    dPointsY[num] += dy;
}

private static int calculateArrowhead1() {

    final double AL = 8.0;
    final double AW = 4.0;

    if (localDump) {
        System.out.println("Before arrowhead: ");
        for (int i = 0; i < 3; i++) {
            System.out.println(String.format("(%f, %f)", dPointsX[i], dPointsY[i]));
        }

        System.out.println(String.format("v21: %f %f; v21-90: %f %f; v21-270: %f %f",
                v21x, v21y, v21x90, v21y90, v21x270, v21y270));
    }

    shiftDPoint(0, v21x * AL + v21x90 * AW, v21y * AL + v21y90 * AW);
    shiftDPoint(2, v21x * AL + v21x270 * AW, v21y * AL + v21y270 * AW);

    if (localDump) {
        System.out.println("After arrowhead: ");
        for (int i = 0; i < 3; i++) {
            System.out.println(String.format("(%f, %f)", dPointsX[i], dPointsY[i]));
        }
    }

    return 3;

}

private static int calculateDiamond1() {

    final double AA = -1.0;
    final double AL = 3.0;
    final double AW = 3.0;

    if (localDump) {
        System.out.println("Before diamond: ");
        for (int i = 0; i < 4; i++) {
            System.out.println(String.format("(%f, %f)", dPointsX[i], dPointsY[i]));
        }

        System.out.println(String.format("v21: %f %f; v21-90: %f %f; v21-270: %f %f",
                v21x, v21y, v21x90, v21y90, v21x270, v21y270));
    }

    shiftDPoint(0, v21x * AA, v21y * AA); // FIXME

    shiftDPoint(1, v21x * (AA + AL) + v21x270 * AW, v21y * (AA + AL) + v21y270 * AW);
    shiftDPoint(3, v21x * (AA + AL) - v21x270 * AW, v21y * (AA + AL) - v21y270 * AW);
    shiftDPoint(2, v21x * (AA + AL * 2), v21y * (AA + AL * 2));

    if (localDump) {
        System.out.println("After diamond: ");
        for (int i = 0; i < 4; i++) {
            System.out.println(String.format("(%f, %f)", dPointsX[i], dPointsY[i]));
        }
    }

    return 4;

}

/**
 * Draws the line end decoration, at the point (x2, y2), according to the line end type
 *
 * @param g           graphic context
 * @param x1          x coordinate for the point 1
 * @param y1          y coordinate for the point 1
 * @param x2          x coordinate for the point 2
 * @param y2          y coordinate for the point 2
 * @param lineEndType decoration type to be drawn
 */
public static void drawLineEndDecoration(Graphics2D g, double x1, double y1, double x2, double y2, LineDecorationType lineEndType) {

    final int CR = 2;
    final int CR2 = 2 * CR;

    GlyphMode glyphMode;

    calculateVectors21(x1, y1, x2, y2);
    setDPointsFromPoint(x2, y2);

    switch (lineEndType) {
        case FILLED_ARROW:
            numPoints = calculateArrowhead1();
            glyphMode = GlyphMode.FILL;
            break;

        case FILLED_DIAMOND:
            numPoints = calculateDiamond1();
            glyphMode = GlyphMode.FILL;
            break;

        case HOLLOW_ARROW:
            numPoints = calculateArrowhead1();
            glyphMode = GlyphMode.HOLLOW;
            break;


        case STANDARD_ARROW:
            numPoints = calculateArrowhead1();
            glyphMode = GlyphMode.STROKE;
            break;

        case FILLED_CIRCLE: // TODO
            g.fillOval((int) x2 - CR, (int) y2 - CR, CR2, CR2);
            return;

        case NONE:
        default:
            numPoints = 0;
            glyphMode = GlyphMode.FILL;

    }

    if (oldDraw) {
        // draw polygon with these points
        setIPointsFromDPoints();
        drawGlyph(g, numPoints, glyphMode);

    } else {
        drawGlyph2D(g, numPoints, glyphMode);
    }


    if (localDump) {
        System.out.println("Drawing " + glyphMode.toString() + " glyph:");
        for (int i = 0; i < numPoints; i++) {
            System.out.println(String.format("(%f, %f)", dPointsX[i], dPointsY[i]));
        }

    }


}

/**
 * Draws the line end decoration, at the point (x2, y2), according to the line end type
 *
 * @param g           graphic context
 * @param p1          the first point
 * @param p2          the second point
 * @param lineEndType
 */
public static void drawLineEndDecoration(Graphics2D g, Point2D p1, Point2D p2, LineDecorationType lineEndType) {
    drawLineEndDecoration(g, p1.getX(), p1.getY(), p2.getX(), p2.getY(), lineEndType);
}


// direction in degrees, starting from 3 == D0
private enum OrthoSegmentDirection {
    NONE, D0, D90, D180, D270
}

private enum GlyphMode {STROKE, FILL, WHITE_FILL, HOLLOW}


}
