/*
*
* Copyright (c) 2013,2014 Institute of Mathematics and Computer Science, University of Latvia (IMCS UL). 
*
* This file is part of DEEK (diagram editor engine kit). 
*
* You can redistribute it and/or modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
*
* You should have received a copy of the GNU General Public License along with DEEK If not, see http://www.gnu.org/licenses/.
*
*/

package lv.lumii.deek.util;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.List;

import lv.lumii.deek.util.Globals;

/**
 * Helper class for checking element hits and some geometric calculations.
 *
 */
public abstract class HitTest {

/**
 * Returns true if the given geometrical ortholine is hit by the given point.
 *
 * @param points
 * @param xp
 * @param yp
 * @return
 */
public static boolean checkOrthoLineHit(List<Point2D.Double> points, double xp,
                                        double yp) {

    Point2D prevPoint, nextPoint;

    prevPoint = points.get(0);

    for (int i = 1; i <= points.size() - 1; i++) {
        nextPoint = points.get(i);
        if (checkOrthoSegmentHit(prevPoint, nextPoint, xp, yp))
            return true;
        prevPoint = nextPoint;
    }
    return false;

}

/**
 * Returns true if the given geometrical polyline is hit by the given point.
 *
 * @param points
 * @param xp
 * @param yp
 * @return
 */
public static boolean checkPolyLineHit(List<Point2D.Double> points, double xp,
                                       double yp) {

    Point2D prevPoint, nextPoint;

    prevPoint = points.get(0);

    for (int i = 1; i < points.size(); i++) {
        nextPoint = points.get(i);
        if (checkFreeSegmentHit(prevPoint, nextPoint, xp, yp))
            return true;
        prevPoint = nextPoint;
    }
    return false;

}

/**
 * Checks whether an orthogonal segment between two points is hit by the given point.
 *
 * @param p1 the segment start
 * @param p2 the segment end
 * @param xp x
 * @param yp y
 * @return true if hit
 */
public static boolean checkOrthoSegmentHit(Point2D p1, Point2D p2,
                                           double xp, double yp) {

    // TODO - add some tolerance for almost H/V segments

    if (p1.getX() == p2.getX()) { // vertical
        if (xp < p1.getX() - Globals.HIT_EPSILON)
            return false;
        if (xp > p1.getX() + Globals.HIT_EPSILON)
            return false;
        if (yp > Math.max(p1.getY(), p2.getY()))
            return false;
        if (yp < Math.min(p1.getY(), p2.getY()))
            return false;
        return true;
    }

    if (p1.getY() == p2.getY()) { // horizontal
        if (yp < p1.getY() - Globals.HIT_EPSILON)
            return false;
        if (yp > p1.getY() + Globals.HIT_EPSILON)
            return false;
        if (xp > Math.max(p1.getX(), p2.getX()))
            return false;
        if (xp < Math.min(p1.getX(), p2.getX()))
            return false;
        return true;
    }

    // segment is not H or V
    return checkFreeSegmentHit(p1, p2, xp, yp);
    // return false;
}

/**
 * Checks whether a segment between two points is hit by the given point.
 *
 * @param p1 the segment start
 * @param p2 the segment end
 * @param xp x
 * @param yp y
 * @return true if hit
 */
public static boolean checkFreeSegmentHit(Point2D p1, Point2D p2,
                                          double xp, double yp) {

    if (Line2D.ptSegDistSq(p1.getX(), p1.getY(), p2.getX(), p2.getY(), xp, yp) < 10)
        return true; // TODO: izdomaat par epsilonu

//		double d = (p2.getX() - p1.getX()) * (yp - p1.getY()) - (p2.getY() - p1.getY()) * (xp - p1.getX());
//		System.out.println("hittest: "+ Double.toString(d));
//		if (Math.abs(d) < 1000) return true; // TODO: izdomaat par epsilonu

    return false;
}

/**
 * Calculates the distance from a point to a polyline.
 *
 * @param points
 * @param point
 * @return
 */
public static double distanceFromPoint(List<Point2D.Double> points, Point2D.Double point) {

    double distance = java.lang.Double.MAX_VALUE;
    double d2;

    Point2D.Double prevPoint, nextPoint;

    prevPoint = points.get(0);

    for (int i = 1; i < points.size(); i++) {
        nextPoint = points.get(i);
        d2 = Line2D.ptSegDist(prevPoint.x, prevPoint.y, nextPoint.x, nextPoint.y, point.x, point.y);

        if (d2 < distance)
            distance = d2;

        prevPoint = nextPoint;
    }

    return distance;
}

/**
 * Calculates the distance from a point to a rectangle.
 *
 * @param r
 * @param point
 * @return
 */
public static double distanceFromPoint(Rectangle2D.Double r, Point2D.Double point) {

    if (r.contains(point.x, point.y))
        return 0;

//		double dx, dy;
//
//		dx = (point.x < r.x ? r.x - point.x : (point.x > r.x + r.width ? point.x - r.x - r.width : 0));
//		dy = (point.y < r.y ? r.y - point.y : (point.y > r.y + r.height ?point.y - r.y - r.height : 0));
//		
//		return dx + dy;

    double d1 = Line2D.ptSegDist(r.x, r.y, r.x + r.width, r.y, point.x, point.y);
    double d2 = Line2D.ptSegDist(r.x + r.width, r.y, r.x + r.width, r.y + r.height, point.x, point.y);
    double d3 = Line2D.ptSegDist(r.x, r.y + r.height, r.x + r.width, r.y + r.height, point.x, point.y);
    double d4 = Line2D.ptSegDist(r.x, r.y, r.x, r.y + r.height, point.x, point.y);

    double d = d1;
    if (d2 < d) d = d2;
    if (d3 < d) d = d3;
    if (d4 < d) d = d4;

    return d;

}

/**
 * Returns new point - the center of the provided rectangle.
 *
 * @param r
 * @return
 */
public static Point2D.Double getCenter(Rectangle2D.Double r) {
    return new Point2D.Double(r.x + r.width / 2, r.y + r.height / 2);
}

}
